from aerosandbox import ExplicitAnalysis
from aerosandbox.geometry import *
from aerosandbox.performance import OperatingPoint
import aerosandbox.library.aerodynamics as aero
import aerosandbox.numpy as np
from aerosandbox.aerodynamics.aero_3D.aero_buildup_submodels.fuselage_aerodynamics_utilities import *
from aerosandbox.library.aerodynamics import transonic
import aerosandbox.library.aerodynamics as aerolib
import copy
from typing import Union, List, Dict, Any
from aerosandbox.aerodynamics.aero_3D.aero_buildup_submodels.softmax_scalefree import softmax_scalefree
from dataclasses import dataclass
from functools import cached_property


class AeroBuildup(ExplicitAnalysis):
    """
    A workbook-style aerodynamics buildup.

    Example usage:

    >>> import aerosandbox as asb
    >>> ab = asb.AeroBuildup(  # This sets up the analysis, but doesn't execute calculation
    >>>     airplane=my_airplane,  # type: asb.Airplane
    >>>     op_point=my_operating_point,  # type: asb.OperatingPoint
    >>>     xyz_ref=[0.1, 0.2, 0.3],  # Moment reference and center of rotation.
    >>> )
    >>> aero = ab.run()  # This executes the actual aero analysis.
    >>> aero_with_stability_derivs = ab.run_with_stability_derivatives()  # Same, but also gets stability derivatives.

    """
    default_analysis_specific_options = {
        Fuselage: dict(
            E_wave_drag=2.5,  # Wave drag efficiency factor

            # Defined by Raymer, "Aircraft Design: A Conceptual Approach", 2nd Ed. Chap. 12.5.9 "Supersonic Parasite Drag".
            # Notated there as "E_WD".
            #
            # Various recommendations:
            #   * For a perfect Sears-Haack body, 1.0
            #   * For a clean aircraft with smooth volume distribution (e.g., BWB), 1.2
            #   * For a "more typical supersonic...", 1.8 - 2.2
            #   * For a "poor supersonic design", 2.5 - 3.0
            #   * The F-15 has E_WD = 2.9.

            nose_fineness_ratio=3,  # Fineness ratio (length / diameter) of the nose section of the fuselage.

            # Impacts wave drag calculations, among other things.
        ),
    }

    def __init__(self,
                 airplane: Airplane,
                 op_point: OperatingPoint,
                 xyz_ref: Union[np.ndarray, List[float]] = None,
                 model_size: str = "small",
                 include_wave_drag: bool = True,
                 ):
        """
        Initializes a new AeroBuildup analysis as an object.

        Note: to run the analysis, you need to first instantiate the object, then call the .run() method.

        Args:
            airplane: The airplane to analyze.

            op_point: The operating point to analyze at. Note that this can be vectorized (i.e., attributes of the OperatingPoint
            object can be arrays, in which case AeroBuildup analysis will be vectorized).

            xyz_ref: The reference point for the aerodynamic forces and moments. This is the point about which the moments are
            taken, and the point at which the forces are applied. Defaults to the airplane's xyz_ref.

            include_wave_drag: Whether to include wave drag in the analysis. Defaults to True.

        Returns: None

        """
        super().__init__()

        ### Set defaults
        if xyz_ref is None:
            xyz_ref = airplane.xyz_ref

        ### Initialize
        self.airplane = airplane
        self.op_point = op_point
        self.xyz_ref = xyz_ref
        self.model_size = model_size
        self.include_wave_drag = include_wave_drag

    def __repr__(self):
        return self.__class__.__name__ + "(\n\t" + "\n\t".join([
            f"airplane={self.airplane}",
            f"op_point={self.op_point}",
            f"xyz_ref={self.xyz_ref}",
        ]) + "\n)"

    @dataclass
    class AeroComponentResults:
        s_ref: float  # Reference area [m^2]
        c_ref: float  # Reference chord [m]
        b_ref: float  # Reference span [m]
        op_point: OperatingPoint
        F_g: List[Union[float, np.ndarray]]  # An [x, y, z] list of forces in geometry axes [N]
        M_g: List[Union[float, np.ndarray]]  # An [x, y, z] list of moments about geometry axes [Nm]
        span_effective: float
        # The effective span of the component's Trefftz-plane wake, used for induced drag calculations. [m]
        oswalds_efficiency: float  # Oswald's efficiency factor [-]

        def __repr__(self):
            F_w = self.F_w
            M_b = self.M_b
            return self.__class__.__name__ + "(\n\t" + "\n\t".join([
                f"L={-F_w[2]},",
                f"Y={F_w[1]},",
                f"D={-F_w[0]},",
                f"l_b={M_b[0]},",
                f"m_b={M_b[1]},",
                f"n_b={M_b[2]},",
                f"span_effective={self.span_effective}, oswalds_efficiency={self.oswalds_efficiency},",
            ]) + "\n)"

        @property
        def F_b(self) -> List[Union[float, np.ndarray]]:
            """
            An [x, y, z] list of forces in body axes [N]
            """
            return self.op_point.convert_axes(*self.F_g, from_axes="geometry", to_axes="body")

        @property
        def F_w(self) -> List[Union[float, np.ndarray]]:
            """
            An [x, y, z] list of forces in wind axes [N]
            """
            return self.op_point.convert_axes(*self.F_g, from_axes="geometry", to_axes="wind")

        @property
        def M_b(self) -> List[Union[float, np.ndarray]]:
            """
            An [x, y, z] list of moments about body axes [Nm]
            """
            return self.op_point.convert_axes(*self.M_g, from_axes="geometry", to_axes="body")

        @property
        def M_w(self) -> List[Union[float, np.ndarray]]:
            """
            An [x, y, z] list of moments about wind axes [Nm]
            """
            return self.op_point.convert_axes(*self.M_g, from_axes="geometry", to_axes="wind")

        @property
        def L(self) -> Union[float, np.ndarray]:
            """
            The lift force [N]. Definitionally, this is in wind axes.
            """
            return -self.F_w[2]

        @property
        def Y(self) -> Union[float, np.ndarray]:
            """
            The side force [N]. Definitionally, this is in wind axes.
            """
            return self.F_w[1]

        @property
        def D(self) -> Union[float, np.ndarray]:
            """
            The drag force [N]. Definitionally, this is in wind axes.
            """
            return -self.F_w[0]

        @property
        def l_b(self) -> Union[float, np.ndarray]:
            """
            The rolling moment [Nm] in body axes. Positive is roll-right.
            """
            return self.M_b[0]

        @property
        def m_b(self) -> Union[float, np.ndarray]:
            """
            The pitching moment [Nm] in body axes. Positive is nose-up.
            """
            return self.M_b[1]

        @property
        def n_b(self) -> Union[float, np.ndarray]:
            """
            The yawing moment [Nm] in body axes. Positive is nose-right.
            """
            return self.M_b[2]

    def run(self) -> Dict[str, Union[Union[float, np.ndarray], List[Union[float, np.ndarray]]]]:
        """
        Computes the aerodynamic forces and moments on the airplane.

        Returns: a dictionary with keys:

            - 'F_g' : an [x, y, z] list of forces in geometry axes [N]
            - 'F_b' : an [x, y, z] list of forces in body axes [N]
            - 'F_w' : an [x, y, z] list of forces in wind axes [N]
            - 'M_g' : an [x, y, z] list of moments about geometry axes [Nm]
            - 'M_b' : an [x, y, z] list of moments about body axes [Nm]
            - 'M_w' : an [x, y, z] list of moments about wind axes [Nm]
            - 'L' : the lift force [N]. Definitionally, this is in wind axes.
            - 'Y' : the side force [N]. This is in wind axes.
            - 'D' : the drag force [N]. Definitionally, this is in wind axes.
            - 'l_b', the rolling moment, in body axes [Nm]. Positive is roll-right.
            - 'm_b', the pitching moment, in body axes [Nm]. Positive is pitch-up.
            - 'n_b', the yawing moment, in body axes [Nm]. Positive is nose-right.
            - 'CL', the lift coefficient [-]. Definitionally, this is in wind axes.
            - 'CY', the sideforce coefficient [-]. This is in wind axes.
            - 'CD', the drag coefficient [-]. Definitionally, this is in wind axes.
            - 'Cl', the rolling coefficient [-], in body axes
            - 'Cm', the pitching coefficient [-], in body axes
            - 'Cn', the yawing coefficient [-], in body axes

            Nondimensional values are nondimensionalized using reference values in the AeroBuildup.airplane object.

            Data types:
                - The "L", "Y", "D", "l_b", "m_b", "n_b", "CL", "CY", "CD", "Cl", "Cm", and "Cn" keys are:

                    - floats if the OperatingPoint object is not vectorized (i.e., if all attributes of OperatingPoint
                    are floats, not arrays).

                    - arrays if the OperatingPoint object is vectorized (i.e., if any attribute of OperatingPoint is an
                    array).

                - The "F_g", "F_b", "F_w", "M_g", "M_b", and "M_w" keys are always lists, which will contain either
                floats or arrays, again depending on whether the OperatingPoint object is vectorized or not.
        """

        ### Compute the forces on each component
        wing_aero_components = [
            self.wing_aerodynamics(
                wing=wing,
                include_induced_drag=False
            )
            for wing in self.airplane.wings
        ]
        fuselage_aero_components = [
            self.fuselage_aerodynamics(
                fuselage=fuse,
                include_induced_drag=False
            )
            for fuse in self.airplane.fuselages
        ]

        aero_components = wing_aero_components + fuselage_aero_components

        ### Sum up the forces
        F_g_total = [
            sum([comp.F_g[i] for comp in aero_components])
            for i in range(3)
        ]
        M_g_total = [
            sum([comp.M_g[i] for comp in aero_components])
            for i in range(3)
        ]

        ##### Add in the induced drag
        Q = self.op_point.dynamic_pressure()

        span_effective_squared = softmax_scalefree([
            comp.span_effective ** 2 * comp.oswalds_efficiency
            for comp in aero_components
        ])

        _, sideforce, lift = self.op_point.convert_axes(
            *F_g_total,
            from_axes="geometry",
            to_axes="wind"
        )

        D_induced = (
                (lift ** 2 + sideforce ** 2) /
                (Q * np.pi * span_effective_squared)
        )

        D_induced_g = self.op_point.convert_axes(
            -D_induced, 0, 0,
            from_axes="wind",
            to_axes="geometry"
        )

        for i in range(3):
            F_g_total[i] += D_induced_g[i]

        ##### Start to assemble the output
        output = {
            "F_g": F_g_total,
            "M_g": M_g_total,
        }

        ##### Add in other metrics
        output["F_b"] = self.op_point.convert_axes(
            *F_g_total,
            from_axes="geometry",
            to_axes="body"
        )
        output["F_w"] = self.op_point.convert_axes(
            *F_g_total,
            from_axes="geometry",
            to_axes="wind"
        )
        output["M_b"] = self.op_point.convert_axes(
            *M_g_total,
            from_axes="geometry",
            to_axes="body"
        )
        output["M_w"] = self.op_point.convert_axes(
            *M_g_total,
            from_axes="geometry",
            to_axes="wind"
        )

        output["L"] = -output["F_w"][2]
        output["Y"] = output["F_w"][1]
        output["D"] = -output["F_w"][0]
        output["l_b"] = output["M_b"][0]
        output["m_b"] = output["M_b"][1]
        output["n_b"] = output["M_b"][2]

        ##### Compute dimensionalization factor
        qS = self.op_point.dynamic_pressure() * self.airplane.s_ref
        c = self.airplane.c_ref
        b = self.airplane.b_ref

        ##### Add nondimensional forces, and nondimensional quantities.
        output["CL"] = output["L"] / qS
        output["CY"] = output["Y"] / qS
        output["CD"] = output["D"] / qS
        output["Cl"] = output["l_b"] / qS / b
        output["Cm"] = output["m_b"] / qS / c
        output["Cn"] = output["n_b"] / qS / b

        ##### Add the component aerodynamics, for reference
        output["wing_aero_components"] = wing_aero_components
        output["fuselage_aero_components"] = fuselage_aero_components

        ##### Add the drag breakdown
        output["D_profile"] = sum([
            comp.D for comp in aero_components
        ])
        output["D_induced"] = D_induced

        return output

    def run_with_stability_derivatives(self,
                                       alpha=True,
                                       beta=True,
                                       p=True,
                                       q=True,
                                       r=True,
                                       ) -> Dict[str, Union[Union[float, np.ndarray], List[Union[float, np.ndarray]]]]:
        """
        Computes the aerodynamic forces and moments on the airplane, and the stability derivatives.

        Arguments essentially determine which stability derivatives are computed. If a stability derivative is not
        needed, leaving it False will speed up the computation.

        Args:

            - alpha (bool): If True, compute the stability derivatives with respect to the angle of attack (alpha).
            - beta (bool): If True, compute the stability derivatives with respect to the sideslip angle (beta).
            - p (bool): If True, compute the stability derivatives with respect to the body-axis roll rate (p).
            - q (bool): If True, compute the stability derivatives with respect to the body-axis pitch rate (q).
            - r (bool): If True, compute the stability derivatives with respect to the body-axis yaw rate (r).

        Returns: a dictionary with keys:

            - 'F_g' : an [x, y, z] list of forces in geometry axes [N]
            - 'F_b' : an [x, y, z] list of forces in body axes [N]
            - 'F_w' : an [x, y, z] list of forces in wind axes [N]
            - 'M_g' : an [x, y, z] list of moments about geometry axes [Nm]
            - 'M_b' : an [x, y, z] list of moments about body axes [Nm]
            - 'M_w' : an [x, y, z] list of moments about wind axes [Nm]
            - 'L'   : the lift force [N]. Definitionally, this is in wind axes.
            - 'Y'   : the side force [N]. This is in wind axes.
            - 'D'   : the drag force [N]. Definitionally, this is in wind axes.
            - 'l_b' : the rolling moment, in body axes [Nm]. Positive is roll-right.
            - 'm_b' : the pitching moment, in body axes [Nm]. Positive is pitch-up.
            - 'n_b' : the yawing moment, in body axes [Nm]. Positive is nose-right.
            - 'CL'  : the lift coefficient [-]. Definitionally, this is in wind axes.
            - 'CY'  : the sideforce coefficient [-]. This is in wind axes.
            - 'CD'  : the drag coefficient [-]. Definitionally, this is in wind axes.
            - 'Cl'  : the rolling coefficient [-], in body axes
            - 'Cm'  : the pitching coefficient [-], in body axes
            - 'Cn'  : the yawing coefficient [-], in body axes

            Along with additional keys, depending on the value of the `alpha`, `beta`, `p`, `q`, and `r` arguments. For
            example, if `alpha=True`, then the following additional keys will be present:

                - 'CLa' : the lift coefficient derivative with respect to alpha [1/rad]
                - 'CDa' : the drag coefficient derivative with respect to alpha [1/rad]
                - 'CYa' : the sideforce coefficient derivative with respect to alpha [1/rad]
                - 'Cla' : the rolling moment coefficient derivative with respect to alpha [1/rad]
                - 'Cma' : the pitching moment coefficient derivative with respect to alpha [1/rad]
                - 'Cna' : the yawing moment coefficient derivative with respect to alpha [1/rad]
                - 'x_np': the neutral point location in the x direction [m]

            Nondimensional values are nondimensionalized using reference values in the AeroBuildup.airplane object.

            Data types:
                - The "L", "Y", "D", "l_b", "m_b", "n_b", "CL", "CY", "CD", "Cl", "Cm", and "Cn" keys are:

                    - floats if the OperatingPoint object is not vectorized (i.e., if all attributes of OperatingPoint
                    are floats, not arrays).

                    - arrays if the OperatingPoint object is vectorized (i.e., if any attribute of OperatingPoint is an
                    array).

                - The "F_g", "F_b", "F_w", "M_g", "M_b", and "M_w" keys are always lists, which will contain either
                floats or arrays, again depending on whether the OperatingPoint object is vectorized or not.

        """
        do_analysis: Dict[str, bool] = {
            "alpha": alpha,
            "beta" : beta,
            "p"    : p,
            "q"    : q,
            "r"    : r,
        }

        abbreviations: Dict[str, str] = {
            "alpha": "a",
            "beta" : "b",
            "p"    : "p",
            "q"    : "q",
            "r"    : "r",
        }
        finite_difference_amounts: Dict[str, float] = {
            "alpha": 0.001,
            "beta" : 0.001,
            "p"    : 0.001 * (2 * self.op_point.velocity) / self.airplane.b_ref,
            "q"    : 0.001 * (2 * self.op_point.velocity) / self.airplane.c_ref,
            "r"    : 0.001 * (2 * self.op_point.velocity) / self.airplane.b_ref,
        }
        scaling_factors: Dict[str, float] = {
            "alpha": np.degrees(1),
            "beta" : np.degrees(1),
            "p"    : (2 * self.op_point.velocity) / self.airplane.b_ref,
            "q"    : (2 * self.op_point.velocity) / self.airplane.c_ref,
            "r"    : (2 * self.op_point.velocity) / self.airplane.b_ref,
        }

        original_op_point = self.op_point

        # Compute the point analysis, which returns a dictionary that we will later add key:value pairs to.
        run_base = self.run()

        # Note for the loops below: here, "derivative numerator" and "... denominator" refer to the quantity being
        # differentiated and the variable of differentiation, respectively. In other words, in the expression df/dx,
        # the "numerator" is f, and the "denominator" is x. I realize that this would make a mathematician cry (as a
        # partial derivative is not a fraction), but the reality is that there seems to be no commonly-accepted name
        # for these terms. (Curiously, this contrasts with integration, where there is an "integrand" and a "variable
        # of integration".)

        for d in do_analysis.keys():
            if not do_analysis[d]:  # Basically, if the parameter from the function input is not True,
                continue  # Skip this run.
                # This way, you can (optionally) speed up this routine if you only need static derivatives,
                # or longitudinal derivatives, etc.

            # These lines make a copy of the original operating point, incremented by the finite difference amount
            # along the variable defined by derivative_denominator.
            incremented_op_point = self.op_point.copy()
            if d == "alpha":
                incremented_op_point.alpha += finite_difference_amounts["alpha"]
            elif d == "beta":
                incremented_op_point.beta += finite_difference_amounts["beta"]
            elif d == "p":
                incremented_op_point.p += finite_difference_amounts["p"]
            elif d == "q":
                incremented_op_point.q += finite_difference_amounts["q"]
            elif d == "r":
                incremented_op_point.r += finite_difference_amounts["r"]
            else:
                raise ValueError(f"Invalid value of d: {d}!")

            aerobuildup_incremented = self.copy()
            aerobuildup_incremented.op_point = incremented_op_point
            run_incremented = aerobuildup_incremented.run()

            for derivative_numerator in [
                "CL",
                "CD",
                "CY",
                "Cl",
                "Cm",
                "Cn",
            ]:
                derivative_name = derivative_numerator + abbreviations[d]  # Gives "CLa"
                run_base[derivative_name] = (
                        (  # Finite-difference out the derivatives
                                run_incremented[derivative_numerator] - run_base[derivative_numerator]
                        ) / finite_difference_amounts[d]
                        * scaling_factors[d]
                )

            ### Try to compute and append neutral point, if possible
            if d == "alpha":
                Cma = run_base["Cma"]
                CLa = np.where(
                    run_base["CLa"] == 0,
                    np.nan,
                    run_base["CLa"],
                )

                run_base["x_np"] = self.xyz_ref[0] - (Cma / CLa * self.airplane.c_ref)

            if d == "beta":
                Cnb = run_base["Cnb"]
                CYb = np.where(
                    run_base["CYb"] == 0,
                    np.nan,
                    run_base["CYb"],
                )

                run_base["x_np_lateral"] = self.xyz_ref[0] - (Cnb / CYb * self.airplane.b_ref)

        return run_base

    def wing_aerodynamics(self,
                          wing: Wing,
                          include_induced_drag: bool = True,
                          ) -> AeroComponentResults:
        """
        Estimates the aerodynamic forces, moments, and derivatives on a wing in isolation.

        Moments are given with the reference at Wing [0, 0, 0].

        Args:

            wing: A Wing object that you wish to analyze.

            op_point: The OperatingPoint that you wish to analyze the fuselage at.

        Returns:

        """
        ##### Alias a few things for convenience
        op_point = self.op_point
        # wing_options = self.get_options(wing) # currently no wing options

        ##### Compute general wing properties
        wing_MAC = wing.mean_aerodynamic_chord()
        wing_taper = wing.taper_ratio()
        wing_sweep = wing.mean_sweep_angle()
        wing_dihedral = wing.mean_dihedral_angle()

        ##### Compute the wing span properties
        sectional_spans = wing.span(
            type="yz",
            include_centerline_distance=False,
            _sectional=True,
        )
        half_span = sum(sectional_spans)
        if len(wing.xsecs) > 0:
            span_inboard_to_YZ_plane = np.inf
            for i in range(len(wing.xsecs)):
                span_inboard_to_YZ_plane = np.minimum(
                    span_inboard_to_YZ_plane,
                    np.abs(wing._compute_xyz_of_WingXSec(
                        i,
                        x_nondim=0.25, z_nondim=0
                    )[1])
                )
        else:
            span_inboard_to_YZ_plane = 0

        ##### Compute the wing area properties
        xsec_chords = [xsec.chord for xsec in wing.xsecs]
        sectional_chords = [
            (inner_chord + outer_chord) / 2
            for inner_chord, outer_chord in zip(
                xsec_chords[1:],
                xsec_chords[:-1]
            )
        ]
        sectional_areas = [
            span * chord
            for span, chord in zip(
                sectional_spans,
                sectional_chords
            )
        ]
        half_area = sum(sectional_areas)
        area_inboard_to_YZ_plane = span_inboard_to_YZ_plane * wing_MAC

        if wing.symmetric:

            span_0_dihedral = 2 * (half_span + span_inboard_to_YZ_plane * 0.5)
            span_90_dihedral = half_span

            area_0_dihedral = 2 * (half_area + area_inboard_to_YZ_plane * 0.5)
            area_90_dihedral = half_area

            dihedral_factor = np.sind(wing_dihedral) ** 2

            span_effective = (
                    span_0_dihedral +
                    (span_90_dihedral - span_0_dihedral) * dihedral_factor
            )

            area_effective = (
                    area_0_dihedral +
                    (area_90_dihedral - area_0_dihedral) * dihedral_factor
            )

        else:
            span_effective = half_span
            area_effective = half_area

        AR_effective = span_effective ** 2 / area_effective

        mach = op_point.mach()
        # mach_normal = mach * np.cosd(sweep)
        AR_3D_factor = aerolib.CL_over_Cl(
            aspect_ratio=AR_effective,
            mach=mach,
            sweep=wing_sweep,
            Cl_is_compressible=True
        )
        oswalds_efficiency = aerolib.oswalds_efficiency(
            taper_ratio=wing_taper,
            aspect_ratio=AR_effective,
            sweep=wing_sweep,
            fuselage_diameter_to_span_ratio=0  # an assumption
        )
        areas = wing.area(_sectional=True)

        aerodynamic_centers = wing.aerodynamic_center(_sectional=True)

        ### Model for the neutral point movement due to lifting-line unsweep near centerline
        # See /studies/AeroBuildup_LL_unsweep_calibration
        a = AR_effective / (AR_effective + 2)
        s = np.radians(wing_sweep)
        t = np.exp(-wing_taper)
        neutral_point_deviation_due_to_unsweep = -(
            ((((3.557726 ** (a ** 2.8443985)) * ((((s * a) + (t * 1.9149417)) + -1.4449639) * s)) + (
                    a + -0.89228547)) * -0.16073418)
        ) * wing_MAC
        aerodynamic_centers = [
            ac + np.array([neutral_point_deviation_due_to_unsweep, 0, 0])
            for ac in aerodynamic_centers
        ]

        xsec_quarter_chords = [
            wing._compute_xyz_of_WingXSec(
                index=i,
                x_nondim=0.25,
                z_nondim=0,
            )
            for i in range(len(wing.xsecs))
        ]

        def compute_section_aerodynamics(
                sect_id: int,
                mirror_across_XZ: bool = False
        ):
            """
            Computes the forces and moments about self.xyz_ref on a given wing section.
            Args:

                sect_id: Wing section id. An int that can be from 0 to len(wing.xsecs) - 2.

                mirror_across_XZ: Boolean. If true, computes the forces and moments for the section that is mirrored across the XZ plane.

            Returns: Forces and moments, in a `(F_g, M_g)` tuple, where `F_g` and `M_g` have the following formats:

                F_g: a [Fx, Fy, Fz] list, given in geometry (`_g`) axes.

                M_g: a [Mx, My, Mz] list, given in geometry (`_g`) axes. Moment reference is `AeroBuildup.xyz_ref`.

            """

            ##### Identify the wing cross sections adjacent to this wing section.
            xsec_a = wing.xsecs[sect_id]
            xsec_b = wing.xsecs[sect_id + 1]

            ##### When linearly interpolating, weight things by the relative chord.
            a_weight = xsec_a.chord / (xsec_a.chord + xsec_b.chord)
            b_weight = xsec_b.chord / (xsec_a.chord + xsec_b.chord)
            mean_chord = (xsec_a.chord + xsec_b.chord) / 2

            ##### Compute the local frame of this section.
            xg_local, yg_local, zg_local = wing._compute_frame_of_section(sect_id)
            xg_local = [xg_local[0], xg_local[1], xg_local[2]]  # convert it to a list
            yg_local = [yg_local[0], yg_local[1], yg_local[2]]  # convert it to a list
            zg_local = [zg_local[0], zg_local[1], zg_local[2]]  # convert it to a list
            if mirror_across_XZ:
                xg_local[1] *= -1
                yg_local[1] *= -1
                zg_local[1] *= -1  # Note: if mirrored, this results in a left-handed coordinate system.

            ##### Compute the moment arm from the section AC
            sect_AC_raw = aerodynamic_centers[sect_id]
            if mirror_across_XZ:
                sect_AC_raw[1] *= -1

            sect_AC = [
                sect_AC_raw[i] - self.xyz_ref[i]
                for i in range(3)
            ]

            ##### Compute the generalized angle of attack, which is the geometric alpha that the wing section "sees".
            vel_vector_g_from_freestream = op_point.convert_axes(  # Points backwards (with relative wind)
                x_from=-op_point.velocity, y_from=0, z_from=0,
                from_axes="wind",
                to_axes="geometry"
            )
            vel_vector_g_from_rotation = np.cross(
                sect_AC,
                op_point.convert_axes(
                    op_point.p, op_point.q, op_point.r,
                    from_axes="body",
                    to_axes="geometry"
                ),
                manual=True
            )
            vel_vector_g = [
                vel_vector_g_from_freestream[i] + vel_vector_g_from_rotation[i]
                for i in range(3)
            ]
            vel_mag_g = np.sqrt(sum(comp ** 2 for comp in vel_vector_g))
            vel_dir_g = [
                vel_vector_g[i] / vel_mag_g
                for i in range(3)
            ]
            vel_dot_x = np.dot(vel_dir_g, xg_local, manual=True)
            vel_dot_z = np.dot(vel_dir_g, zg_local, manual=True)

            # alpha_generalized = 90 - np.arccosd(np.clip(vel_dot_z, -1, 1)) # In range (-90 to 90)
            alpha_generalized = np.where(
                vel_dot_x > 0,
                90 - np.arccosd(np.clip(vel_dot_z, -1, 1)),  # In range (-90 to 90)
                90 + np.arccosd(np.clip(vel_dot_z, -1, 1))  # In range (90 to 270)
            )

            ##### Compute the effective generalized angle of attack, which roughly accounts for self-downwash
            # effects (e.g., finite-wing effects on lift curve slope). Despite this being a tuned heuristic,
            # it is surprisingly accurate! (<20% lift coefficient error against wind tunnel experiment, even at as
            # low as AR = 0.5.)
            alpha_generalized_effective = (
                    alpha_generalized -
                    (1 - AR_3D_factor ** 0.8) * np.sind(2 * alpha_generalized) / 2 * (180 / np.pi)
                # TODO: "center" this scaling around alpha = alpha_{airfoil, Cl=0}, not around alpha = 0.
                # TODO Can estimate airfoil's alpha_{Cl=0} by camber + thin airfoil theory + viscous decambering knockdown.
            )  # Models finite-wing increase in alpha_{CL_max}.

            ##### Compute sweep angle
            xsec_a_quarter_chord = xsec_quarter_chords[sect_id]
            xsec_b_quarter_chord = xsec_quarter_chords[sect_id + 1]
            quarter_chord_vector_g = xsec_b_quarter_chord - xsec_a_quarter_chord
            quarter_chord_dir_g = quarter_chord_vector_g / np.linalg.norm(quarter_chord_vector_g)
            quarter_chord_dir_g = [  # Convert to list
                quarter_chord_dir_g[i]
                for i in range(3)
            ]

            vel_dir_dot_quarter_chord_dir = np.dot(
                vel_dir_g,
                quarter_chord_dir_g,
                manual=True
            )

            sweep_rad = np.arcsin(vel_dir_dot_quarter_chord_dir)

            ##### Compute Reynolds numbers
            Re_a = op_point.reynolds(xsec_a.chord)
            Re_b = op_point.reynolds(xsec_b.chord)

            ##### Compute Mach numbers
            mach_normal = mach * np.cos(sweep_rad)

            ##### Compute effective alpha due to control surface deflections
            symmetry_treated_control_surfaces: List[ControlSurface] = []

            for surf in xsec_a.control_surfaces:
                if mirror_across_XZ and not surf.symmetric:
                    surf = surf.copy()
                    surf.deflection = -surf.deflection
                symmetry_treated_control_surfaces.append(surf)

            ##### Compute sectional lift at cross-sections using lookup functions. Merge them linearly to get section CL.
            kwargs = dict(
                alpha=alpha_generalized_effective,
                mach=mach_normal if self.include_wave_drag else 0.,
                control_surfaces=symmetry_treated_control_surfaces,
                model_size=self.model_size,
            )

            xsec_a_airfoil_aero = xsec_a.airfoil.get_aero_from_neuralfoil(
                Re=Re_a,
                **kwargs
            )
            xsec_b_airfoil_aero = xsec_b.airfoil.get_aero_from_neuralfoil(
                Re=Re_b,
                **kwargs
            )

            xsec_a_Cl = xsec_a_airfoil_aero["CL"]
            xsec_b_Cl = xsec_b_airfoil_aero["CL"]
            sect_CL = (
                              xsec_a_Cl * a_weight +
                              xsec_b_Cl * b_weight
                      ) * AR_3D_factor ** 0.2  # Models slight decrease in finite-wing CL_max.

            ##### Compute sectional drag at cross-sections using lookup functions. Merge them linearly to get section CD.
            xsec_a_Cdp = xsec_a_airfoil_aero["CD"]
            xsec_b_Cdp = xsec_b_airfoil_aero["CD"]
            sect_CDp = (
                (
                        xsec_a_Cdp * a_weight +
                        xsec_b_Cdp * b_weight
                )
            )

            ##### Compute sectional moment at cross-sections using lookup functions. Merge them linearly to get section CM.
            xsec_a_Cm = xsec_a_airfoil_aero["CM"]
            xsec_b_Cm = xsec_b_airfoil_aero["CM"]
            sect_CM = (
                    xsec_a_Cm * a_weight +
                    xsec_b_Cm * b_weight
            )

            ##### Compute induced drag from local CL and full-wing properties (AR, e)
            if include_induced_drag:
                sect_CDi = (
                        sect_CL ** 2 / (np.pi * AR_effective * oswalds_efficiency)
                )

                sect_CD = sect_CDp + sect_CDi
            else:
                sect_CD = sect_CDp

            ##### Go to dimensional quantities using the area.
            area = areas[sect_id]
            q_local = 0.5 * op_point.atmosphere.density() * vel_mag_g ** 2
            sect_L = q_local * area * sect_CL
            sect_D = q_local * area * sect_CD
            sect_M = q_local * area * sect_CM * mean_chord

            ##### Compute the direction of the lift by projecting the section's normal vector into the plane orthogonal to the local freestream.
            L_direction_g_unnormalized = [
                zg_local[i] - vel_dot_z * vel_dir_g[i]
                for i in range(3)
            ]
            L_direction_g_unnormalized = [  # Handles the 90 degree to 270 degree cases
                np.where(
                    vel_dot_x > 0,
                    L_direction_g_unnormalized[i],
                    -1 * L_direction_g_unnormalized[i],
                )
                for i in range(3)
            ]
            L_direction_g_mag = np.sqrt(sum(comp ** 2 for comp in L_direction_g_unnormalized))
            L_direction_g = [
                L_direction_g_unnormalized[i] / L_direction_g_mag
                for i in range(3)
            ]

            ##### Compute the direction of the drag by aligning the drag vector with the freestream vector.
            D_direction_g = vel_dir_g

            ##### Compute the force vector in geometry axes.
            sect_F_g = [
                sect_L * L_direction_g[i] + sect_D * D_direction_g[i]
                for i in range(3)
            ]

            ##### Compute the moment vector in geometry axes.
            M_g_lift = np.cross(
                sect_AC,
                sect_F_g,
                manual=True
            )
            M_direction_g = np.cross(L_direction_g, D_direction_g, manual=True)
            M_g_pitching_moment = [
                M_direction_g[i] * sect_M
                for i in range(3)
            ]
            sect_M_g = [
                M_g_lift[i] + M_g_pitching_moment[i]
                for i in range(3)
            ]

            return sect_F_g, sect_M_g

        ##### Iterate through all sections and add up all forces/moments.
        F_g = [0., 0., 0.]
        M_g = [0., 0., 0.]

        for sect_id in range(len(wing.xsecs) - 1):
            sect_F_g, sect_M_g = compute_section_aerodynamics(sect_id=sect_id)

            for i in range(3):
                F_g[i] += sect_F_g[i]
                M_g[i] += sect_M_g[i]

            if wing.symmetric:
                sect_F_g, sect_M_g = compute_section_aerodynamics(sect_id=sect_id, mirror_across_XZ=True)

                for i in range(3):
                    F_g[i] += sect_F_g[i]
                    M_g[i] += sect_M_g[i]

        return self.AeroComponentResults(
            s_ref=self.airplane.s_ref,
            c_ref=self.airplane.c_ref,
            b_ref=self.airplane.b_ref,
            op_point=op_point,
            F_g=F_g,
            M_g=M_g,
            span_effective=span_effective,
            oswalds_efficiency=oswalds_efficiency
        )

    def fuselage_aerodynamics(self,
                              fuselage: Fuselage,
                              include_induced_drag: bool = True
                              ) -> AeroComponentResults:
        """
        Estimates the aerodynamic forces, moments, and derivatives on a fuselage in isolation.

        Assumes:
            * The fuselage is a body of revolution aligned with the x_b axis.
            * The angle between the nose and the freestream is less than 90 degrees.

        Moments are given with the reference at Fuselage [0, 0, 0].

        Uses methods from Jorgensen, Leland Howard. "Prediction of Static Aerodynamic Characteristics for Slender Bodies
        Alone and with Lifting Surfaces to Very High Angles of Attack". NASA TR R-474. 1977.

        Args:

            fuselage: A Fuselage object that you wish to analyze.

        Returns:

        """
        ##### Alias a few things for convenience
        op_point = self.op_point
        length = fuselage.length()
        Re = op_point.reynolds(reference_length=length)
        mach = op_point.mach()
        fuse_options = self.get_options(fuselage)

        ##### Compute general fuselage properties
        q = op_point.dynamic_pressure()
        eta = jorgensen_eta(fuselage.fineness_ratio())

        span_effective = softmax_scalefree(
            [
                xsec.width
                for xsec in fuselage.xsecs
            ] + [
                xsec.height
                for xsec in fuselage.xsecs
            ],
        )

        ##### Initialize storage for total forces and moments, in geometry axes.
        F_g = [0., 0., 0.]
        M_g = [0., 0., 0.]

        ##### Compute the inviscid aerodynamics using slender body theory
        xsec_areas = [
            xsec.xsec_area()
            for xsec in fuselage.xsecs
        ]

        sect_xyz_a = [
            xsec.xyz_c
            for xsec in fuselage.xsecs[:-1]
        ]

        sect_xyz_b = [
            xsec.xyz_c
            for xsec in fuselage.xsecs[1:]
        ]

        sect_xyz_midpoints = [
            [(xyz_a[i] + xyz_b[i]) / 2 for i in range(3)]
            for xyz_a, xyz_b in zip(sect_xyz_a, sect_xyz_b)
        ]

        sect_lengths = [
            (sum((xyz_a[i] - xyz_b[i]) ** 2 for i in range(3))) ** 0.5
            for xyz_a, xyz_b in zip(sect_xyz_a, sect_xyz_b)
        ]

        sect_directions = [
            [np.where(
                sect_lengths[i] != 0,
                (xyz_b[j] - xyz_a[j]) / (sect_lengths[i] + 1e-100),
                1 if j == 0 else 0  # Default to [1, 0, 0]
            )
                for j in range(3)
            ]
            for i, (xyz_a, xyz_b) in enumerate(zip(sect_xyz_a, sect_xyz_b))
        ]

        sect_areas = [
            (area_a + area_b + (area_a * area_b + 1e-100) ** 0.5) / 3
            for area_a, area_b in zip(xsec_areas[:-1], xsec_areas[1:])
        ]

        vel_direction_g = op_point.convert_axes(-1, 0, 0, from_axes="wind", to_axes="geometry")

        sin_local_alpha_force_direction = [
            [
                np.dot(s, vel_direction_g, manual=True) * vel_direction_g[i] - s[i]
                for i in range(3)
            ]
            for s in sect_directions
        ]

        rho_V_squared = op_point.atmosphere.density() * op_point.velocity ** 2

        sin_local_alpha_moment_direction = [
            np.cross(
                vel_direction_g,
                sect_direction,
                manual=True
            )
            for sect_direction in sect_directions
        ]

        # Drela, Flight Vehicle Aerodynamics Eq. 6.77
        lift_force_at_nose = [
            rho_V_squared
            * xsec_areas[-1]
            * sin_local_alpha_force_direction[-1][i]
            for i in range(3)
        ]

        # Drela, Flight Vehicle Aerodynamics Eq. 6.78
        moment_at_nose_due_to_open_tail = [
            -1 * rho_V_squared
            * sum(sect_lengths)
            * xsec_areas[-1]
            * sin_local_alpha_moment_direction[-1][i]
            for i in range(3)
        ]
        moment_at_nose_due_to_shape = [
            rho_V_squared
            * sum(
                area * moment_direction[i] * length
                for area, moment_direction, length in zip(
                    sect_areas,
                    sin_local_alpha_moment_direction,
                    sect_lengths
                )
            )
            for i in range(3)
        ]

        lift_moment_arm = [
            fuselage.xsecs[0].xyz_c[i] - self.xyz_ref[i]
            for i in range(3)
        ]
        moment_due_to_lift_force = np.cross(
            lift_moment_arm,
            lift_force_at_nose,
            manual=True
        )

        ### Total the invsicid forces and moments
        for i in range(3):
            F_g[i] += lift_force_at_nose[i]
            M_g[i] += moment_at_nose_due_to_open_tail[i] + moment_at_nose_due_to_shape[i] + moment_due_to_lift_force[i]

        ##### Now, need to add in viscous aerodynamics from profile drag sources
        ### Base Drag
        base_drag_coefficient = fuselage_base_drag_coefficient(mach=mach)
        drag_base = base_drag_coefficient * fuselage.area_base() * q

        ### Skin friction drag
        form_factor = fuselage_form_factor(
            fineness_ratio=fuselage.fineness_ratio(),
            ratio_of_corner_radius_to_body_width=0.5
        )
        C_f_ideal = (
                        # From the same study as the `fuselage_form_factor` function above. This is done on purpose
                        # as the form factor in this particular paper is a fit that correlates best using this precise
                        # definition of C_f_ideal.
                            3.46 * np.log10(Re) - 5.6
                    ) ** -2
        C_f = C_f_ideal * form_factor
        drag_skin = C_f * fuselage.area_wetted() * q

        ### Wave drag
        S_ref = 1  # Does not matter here, just for accounting.

        if self.include_wave_drag:
            sears_haack_drag_area = transonic.sears_haack_drag_from_volume(
                volume=fuselage.volume(),
                length=length
            )  # Units of area
            sears_haack_C_D_wave = sears_haack_drag_area / S_ref

            C_D_wave = transonic.approximate_CD_wave(
                mach=mach,
                mach_crit=critical_mach(
                    fineness_ratio_nose=fuse_options["nose_fineness_ratio"]
                ),
                CD_wave_at_fully_supersonic=fuse_options["E_wave_drag"] * sears_haack_C_D_wave,
            )
        else:
            C_D_wave = 0

        drag_wave = C_D_wave * q * S_ref

        ### Sum up the profile drag
        drag_profile = drag_base + drag_skin + drag_wave

        drag_profile_g = op_point.convert_axes(
            -drag_profile, 0, 0,
            from_axes="wind",
            to_axes="geometry"
        )

        drag_moment_arm = [
            (fuselage.xsecs[0].xyz_c[i] + fuselage.xsecs[-1].xyz_c[i]) / 2 - self.xyz_ref[i]
            for i in range(3)
        ]

        moment_due_to_drag_profile = np.cross(
            drag_moment_arm,
            drag_profile_g,
            manual=True
        )

        for i in range(3):
            F_g[i] += drag_profile_g[i]
            M_g[i] += moment_due_to_drag_profile[i]

        ##### Now, we need to add in the viscous aerodynamics from crossflow sources
        sin_generalized_alphas = [
            sum(comp ** 2 for comp in s) ** 0.5
            for s in sin_local_alpha_force_direction
        ]
        mean_aerodynamic_radii = [
            (area / np.pi + 1e-100) ** 0.5
            for area in sect_areas
        ]

        Re_n_sect = [
            s * op_point.reynolds(
                reference_length=2 * radius
            )
            for s, radius in zip(sin_generalized_alphas, mean_aerodynamic_radii)
        ]
        mach_n_sect = [
            s * mach
            for s in sin_generalized_alphas
        ]

        C_d_n_sect = [
            np.where(
                Re_n_sect[i] != 0,
                aerolib.Cd_cylinder(
                    Re_D=Re_n_sect[i],
                    mach=mach_n_sect[i]
                ),  # Replace with 1.20 from Jorgensen Table 1 if this isn't working well
                0,
            )
            for i in range(len(fuselage.xsecs) - 1)
        ]

        vel_dot_x = [
            np.dot(
                vel_direction_g,
                sect_directions[i],
                manual=True
            )
            for i in range(len(fuselage.xsecs) - 1)
        ]
        normal_directions_g_unnormalized = [
            [
                vel_direction_g[j] - vel_dot_x[i] * sect_directions[i][j]
                for j in range(3)
            ]
            for i in range(len(fuselage.xsecs) - 1)
        ]
        for i in range(len(fuselage.xsecs) - 1):
            normal_directions_g_unnormalized[i][2] += 1e-100  # Hack to avoid divide-by-zero in 0-AoA case

        normal_directions_g_mag = [
            (sum(comp ** 2 for comp in n) + 1e-100) ** 0.5
            for n in normal_directions_g_unnormalized
        ]
        normal_directions_g = [
            [
                normal_directions_g_unnormalized[i][j] / normal_directions_g_mag[i]
                for j in range(3)
            ]
            for i in range(len(fuselage.xsecs) - 1)
        ]

        lift_viscous_crossflow = [
            [
                sect_lengths[i]
                * rho_V_squared
                * eta
                * C_d_n_sect[i]
                * normal_directions_g[i][j] * sum(c ** 2 for c in sin_local_alpha_force_direction[i])
                * mean_aerodynamic_radii[i]
                for j in range(3)
            ]
            for i in range(len(fuselage.xsecs) - 1)
        ]

        moment_viscous_crossflow = [
            np.cross(
                [sect_xyz_midpoints[i][j] - self.xyz_ref[j] for j in range(3)],
                lift_viscous_crossflow[i],
                manual=True
            )
            for i in range(len(fuselage.xsecs) - 1)
        ]

        F_g_viscous_crossflow = [
            sum(lift_viscous_crossflow[i][j] for i in range(len(fuselage.xsecs) - 1))
            for j in range(3)
        ]

        M_g_viscous_crossflow = [
            sum(moment_viscous_crossflow[i][j] for i in range(len(fuselage.xsecs) - 1))
            for j in range(3)
        ]

        for i in range(3):
            F_g[i] += F_g_viscous_crossflow[i]
            M_g[i] += M_g_viscous_crossflow[i]

        ### Compute the induced drag, if relevant
        if include_induced_drag:
            _, sideforce, lift = op_point.convert_axes(
                *F_g,
                from_axes="geometry",
                to_axes="wind"
            )

            D_induced = (
                    (lift ** 2 + sideforce ** 2) /
                    (op_point.dynamic_pressure() * np.pi * span_effective ** 2)
            )

            D_induced_g = op_point.convert_axes(
                -D_induced, 0, 0,
                from_axes="wind",
                to_axes="geometry",
            )

            for i in range(3):
                F_g[i] += D_induced_g[i]

        return self.AeroComponentResults(
            s_ref=self.airplane.s_ref,
            c_ref=self.airplane.c_ref,
            b_ref=self.airplane.b_ref,
            op_point=op_point,
            F_g=F_g,
            M_g=M_g,
            span_effective=span_effective,
            oswalds_efficiency=0.95,
        )


if __name__ == '__main__':
    from aerosandbox.aerodynamics.aero_3D.test_aero_3D.geometries.conventional import airplane
    import matplotlib.pyplot as plt
    import aerosandbox.tools.pretty_plots as p

    aero = AeroBuildup(
        airplane=airplane,
        op_point=OperatingPoint(alpha=0, beta=1),
    ).run()

    fig, ax = plt.subplots(2, 2)
    alpha = np.linspace(-20, 20, 1000)
    aero = AeroBuildup(
        airplane=airplane,
        op_point=OperatingPoint(
            velocity=10,
            alpha=alpha,
            beta=0
        ),
    ).run()

    plt.sca(ax[0, 0])
    plt.plot(alpha, aero["CL"])
    plt.xlabel(r"$\alpha$ [deg]")
    plt.ylabel(r"$C_L$")
    p.set_ticks(5, 1, 0.5, 0.1)

    plt.sca(ax[0, 1])
    plt.plot(alpha, aero["CD"])
    plt.xlabel(r"$\alpha$ [deg]")
    plt.ylabel(r"$C_D$")
    p.set_ticks(5, 1, 0.05, 0.01)
    plt.ylim(bottom=0)

    plt.sca(ax[1, 0])
    plt.plot(alpha, aero["Cm"])
    plt.xlabel(r"$\alpha$ [deg]")
    plt.ylabel(r"$C_m$")
    p.set_ticks(5, 1, 0.5, 0.1)

    plt.sca(ax[1, 1])
    plt.plot(alpha, aero["CL"] / aero["CD"])
    plt.xlabel(r"$\alpha$ [deg]")
    plt.ylabel(r"$C_L/C_D$")
    p.set_ticks(5, 1, 10, 2)

    p.show_plot(
        "`asb.AeroBuildup` Aircraft Aerodynamics"
    )

    Beta, Alpha = np.meshgrid(np.linspace(-90, 90, 200), np.linspace(-90, 90, 200))
    aero = AeroBuildup(
        airplane=airplane,
        op_point=OperatingPoint(
            velocity=10,
            alpha=Alpha.flatten(),
            beta=Beta.flatten()
        ),
    ).run()


    def show():
        p.set_ticks(15, 5, 15, 5)
        p.equal()
        p.show_plot(
            "`asb.AeroBuildup` Aircraft Aerodynamics",
            r"Sideslip angle $\beta$ [deg]",
            r"Angle of Attack $\alpha$ [deg]"
        )


    fig, ax = plt.subplots(figsize=(6, 5))
    p.contour(
        Beta, Alpha, aero["CL"].reshape(Alpha.shape),
        colorbar_label="Lift Coefficient $C_L$ [-]",
        linelabels_format=lambda x: f"{x:.2f}",
        linelabels_fontsize=7,
        cmap="RdBu",
        alpha=0.6
    )
    plt.clim(*np.array([-1, 1]) * np.max(np.abs(aero["CL"])))
    show()

    fig, ax = plt.subplots(figsize=(6, 5))
    p.contour(
        Beta, Alpha, aero["CD"].reshape(Alpha.shape),
        colorbar_label="Drag Coefficient $C_D$ [-]",
        linelabels_format=lambda x: f"{x:.2f}",
        linelabels_fontsize=7,
        z_log_scale=True,
        cmap="YlOrRd",
        alpha=0.6
    )
    show()

    fig, ax = plt.subplots(figsize=(6, 5))
    p.contour(
        Beta, Alpha, (aero["CL"] / aero["CD"]).reshape(Alpha.shape),
        levels=15,
        colorbar_label="$C_L / C_D$ [-]",
        linelabels_format=lambda x: f"{x:.0f}",
        linelabels_fontsize=7,
        cmap="RdBu",
        alpha=0.6
    )
    plt.clim(*np.array([-1, 1]) * np.max(np.abs(aero["CL"] / aero["CD"])))
    show()



from aerosandbox.common import ExplicitAnalysis
import aerosandbox.numpy as np
import subprocess
from pathlib import Path
from aerosandbox.geometry import Airplane, Wing, WingXSec, Fuselage, ControlSurface
from aerosandbox.performance import OperatingPoint
from typing import Union, List, Dict, Any
import copy
import tempfile
import warnings


class AVL(ExplicitAnalysis):
    """

    An interface to AVL, a 3D vortex lattice aerodynamics code developed by Mark Drela at MIT.

    Requires AVL to be on your computer; AVL is available here: https://web.mit.edu/drela/Public/web/avl/

    It is recommended (but not required) that you add AVL to your system PATH environment variable such that it can
    be called with the command `avl`. If this is not the case, you need to specify the path to your AVL
    executable using the `avl_command` argument of the constructor.

    Usage example:

        >>> avl = asb.AVL(
        >>>     airplane=my_airplane,
        >>>     op_point=asb.OperatingPoint(
        >>>         velocity=100, # m/s
        >>>         alpha=5, # deg
        >>>         beta=4, # deg
        >>>         p=0.01, # rad/sec
        >>>         q=0.02, # rad/sec
        >>>         r=0.03, # rad/sec
        >>>     )
        >>> )
        >>> outputs = avl.run()

    """
    default_analysis_specific_options = {
        Airplane: dict(
            profile_drag_coefficient=0
        ),
        Wing    : dict(
            wing_level_spanwise_spacing=True,
            spanwise_resolution=12,
            spanwise_spacing="cosine",
            chordwise_resolution=12,
            chordwise_spacing="cosine",
            component=None,  # This is an int
            no_wake=False,
            no_alpha_beta=False,
            no_load=False,
            drag_polar=dict(
                CL1=0,
                CD1=0,
                CL2=0,
                CD2=0,
                CL3=0,
                CD3=0,
            ),
        ),
        WingXSec: dict(
            spanwise_resolution=12,
            spanwise_spacing="cosine",
            cl_alpha_factor=None,  # This is a float
            drag_polar=dict(
                CL1=0,
                CD1=0,
                CL2=0,
                CD2=0,
                CL3=0,
                CD3=0,
            )
        ),
        Fuselage: dict(
            panel_resolution=24,
            panel_spacing="cosine"
        )
    }

    AVL_spacing_parameters = {
        "uniform": 0,
        "cosine" : 1,
        "sine"   : 2,
        "-sine"  : -2,
        "equal"  : 0,  # "uniform" is preferred
    }

    def __init__(self,
                 airplane: Airplane,
                 op_point: OperatingPoint,
                 xyz_ref: List[float] = None,
                 avl_command: str = "avl",
                 verbose: bool = False,
                 timeout: Union[float, int, None] = 5,
                 working_directory: str = None,
                 ground_effect: bool = False,
                 ground_effect_height: float = 0
                 ):
        """
        Interface to AVL.

        Args:

            airplane: The airplane object you wish to analyze.

            op_point: The operating point you wish to analyze at.

            avl_command: The command-line argument to call AVL.

                * If AVL is on your system PATH, then you can just leave this as "avl".

                * If AVL is not on your system PATH, then you should provide a filepath to the AVL executable.

                Note that AVL is not on your PATH by default. To tell if AVL is on your system PATH, open up a
                terminal and type "avl".

                    * If the AVL menu appears, it's on your PATH.

                    * If you get something like "'avl' is not recognized as an internal or external command..." or
                    "Command 'avl' not found, did you mean...", then it is not on your PATH and you'll need to
                    specify the location of your AVL executable as a string.

                To add AVL to your path, modify your system's environment variables. (Google how to do this for your OS.)

            verbose: Controls whether or not AVL output is printed to command line.

            timeout: Controls how long any individual AVL run is allowed to run before the
            process is killed. Given in units of seconds. To disable timeout, set this to None.

            working_directory: Controls which working directory is used for the AVL input and output files. By
            default, this is set to a TemporaryDirectory that is deleted after the run. However, you can set it to
            somewhere local for debugging purposes.
        """
        super().__init__()

        ### Set defaults
        if xyz_ref is None:
            xyz_ref = airplane.xyz_ref

        ### Initialize
        self.airplane = airplane
        self.op_point = op_point
        self.xyz_ref = xyz_ref
        self.avl_command = avl_command
        self.verbose = verbose
        self.timeout = timeout
        self.working_directory = working_directory
        self.ground_effect = ground_effect
        self.ground_effect_height = ground_effect_height

    def __repr__(self):
        return self.__class__.__name__ + "(\n\t" + "\n\t".join([
            f"airplane={self.airplane}",
            f"op_point={self.op_point}",
            f"xyz_ref={self.xyz_ref}",
        ]) + "\n)"

    def open_interactive(self) -> None:
        """
        Opens a new terminal window and runs AVL interactively. This is useful for detailed analysis or debugging.

        Returns: None
        """
        with tempfile.TemporaryDirectory() as directory:
            directory = Path(directory)

            ### Alternatively, work in another directory:
            if self.working_directory is not None:
                directory = Path(self.working_directory)  # For debugging

            ### Handle the airplane file
            airplane_file = "airplane.avl"
            self.write_avl(directory / airplane_file)

            ### Open up AVL
            import sys, os
            if sys.platform == "win32":
                # Run AVL
                print("Running AVL interactively in a new window, quit it to continue...")

                command = f'cmd /k "{self.avl_command} {airplane_file}"'

                process = subprocess.Popen(
                    command,
                    cwd=directory,
                    creationflags=subprocess.CREATE_NEW_CONSOLE,
                )
                process.wait()

            else:
                raise NotImplementedError(
                    "Ability to auto-launch interactive AVL sessions isn't yet implemented for non-Windows OSes."
                )

    def run(self,
            run_command: str = None,
            ) -> Dict[str, float]:
        """
        Private function to run AVL.

        Args: run_command: A string with any AVL keystroke inputs that you'd like. By default, you start off within the OPER
        menu. All of the inputs indicated in the constructor have been set already, but you can override them here (
        for this run only) if you want.

        Returns: A dictionary containing all of your results.

        """
        with tempfile.TemporaryDirectory() as directory:
            directory = Path(directory)

            ### Alternatively, work in another directory:
            if self.working_directory is not None:
                directory = Path(self.working_directory)  # For debugging

            # Designate an intermediate file for file I/O
            output_filename = "output.txt"

            # Handle the airplane file
            airplane_file = "airplane.avl"
            self.write_avl(directory / airplane_file)

            # Handle the run file
            keystroke_file_contents = self._default_keystroke_file_contents()
            if run_command is not None:
                keystroke_file_contents += [run_command]
            keystroke_file_contents += [
                "x",
                "st",
                f"{output_filename}",
                "o",
                "",
                "",
                "quit"
            ]

            keystrokes = "\n".join(keystroke_file_contents)

            command = f'{self.avl_command} {airplane_file}'

            ### Execute
            try:
                proc = subprocess.Popen(
                    command,
                    cwd=directory,
                    stdin=subprocess.PIPE,
                    stdout=None if self.verbose else subprocess.DEVNULL,
                    stderr=None if self.verbose else subprocess.DEVNULL,
                    text=True,
                    # shell=True,
                    # timeout=self.timeout,
                    # check=True
                )
                outs, errs = proc.communicate(
                    input=keystrokes,
                    timeout=self.timeout
                )
                return_code = proc.poll()

            except subprocess.TimeoutExpired:
                proc.kill()
                outs, errs = proc.communicate()

                warnings.warn(
                    "AVL run timed out!\n"
                    "If this was not expected, try increasing the `timeout` parameter\n"
                    "when you create this AeroSandbox AVL instance.",
                    stacklevel=2
                )

            ##### Parse the output file
            # Read the file
            try:
                with open(directory / output_filename, "r") as f:
                    output_data = f.read()
            except FileNotFoundError:
                raise FileNotFoundError(
                    "It appears AVL didn't produce an output file, probably because it crashed.\n"
                    "To troubleshoot, try some combination of the following:\n"
                    "\t - In the AVL constructor, verify that either AVL is on PATH or that the `avl_command` parameter is set.\n"
                    "\t - In the AVL constructor, run with `verbose=True`.\n"
                    "\t - In the AVL constructor, set the `working_directory` parameter to a known folder to see the AVL input and output files.\n"
                    "\t - In the AVL constructor, set the `timeout` parameter to a large number to see if AVL is just taking a long time to run.\n"
                    "\t - On Windows, use `avl.open_interactive()` to run AVL interactively in a new window.\n"
                )

            res = self.parse_unformatted_data_output(output_data, data_identifier=" =", overwrite=False)

            ##### Clean up results
            for key_to_lowerize in ["Alpha", "Beta", "Mach"]:
                res[key_to_lowerize.lower()] = res.pop(key_to_lowerize)

            for key in list(res.keys()):
                if "tot" in key:
                    res[key.replace("tot", "")] = res.pop(key)

            ##### Add in missing useful results
            q = self.op_point.dynamic_pressure()
            S = self.airplane.s_ref
            b = self.airplane.b_ref
            c = self.airplane.c_ref

            res["p"] = res["pb/2V"] * (2 * self.op_point.velocity / b)
            res["q"] = res["qc/2V"] * (2 * self.op_point.velocity / c)
            res["r"] = res["rb/2V"] * (2 * self.op_point.velocity / b)
            res["L"] = q * S * res["CL"]
            res["Y"] = q * S * res["CY"]
            res["D"] = q * S * res["CD"]
            res["l_b"] = q * S * b * res["Cl"]
            res["m_b"] = q * S * c * res["Cm"]
            res["n_b"] = q * S * b * res["Cn"]
            try:
                res["Clb Cnr / Clr Cnb"] = res["Clb"] * res["Cnr"] / (res["Clr"] * res["Cnb"])
            except ZeroDivisionError:
                res["Clb Cnr / Clr Cnb"] = np.nan

            res["F_w"] = [
                -res["D"], res["Y"], -res["L"]
            ]
            res["F_b"] = self.op_point.convert_axes(*res["F_w"], from_axes="wind", to_axes="body")
            res["F_g"] = self.op_point.convert_axes(*res["F_b"], from_axes="body", to_axes="geometry")
            res["M_b"] = [
                res["l_b"], res["m_b"], res["n_b"]
            ]
            res["M_g"] = self.op_point.convert_axes(*res["M_b"], from_axes="body", to_axes="geometry")
            res["M_w"] = self.op_point.convert_axes(*res["M_b"], from_axes="body", to_axes="wind")

            return res

    def _default_keystroke_file_contents(self) -> List[str]:

        run_file_contents = []

        # Disable graphics
        run_file_contents += [
            "plop",
            "g",
            "",
        ]

        # Enter oper mode
        run_file_contents += [
            "oper",
        ]

        # Direct p, q, r to be in body axes, to match ASB convention
        run_file_contents += [
            "o",
            "r",
            "",
        ]

        # Set parameters
        run_file_contents += [
            "m",
            f"mn {float(self.op_point.mach())}",
            f"v {float(self.op_point.velocity)}",
            f"d {float(self.op_point.atmosphere.density())}",
            "g 9.81",
            ""
        ]

        # Set analysis state
        p_bar = self.op_point.p * self.airplane.b_ref / (2 * self.op_point.velocity)
        q_bar = self.op_point.q * self.airplane.c_ref / (2 * self.op_point.velocity)
        r_bar = self.op_point.r * self.airplane.b_ref / (2 * self.op_point.velocity)

        run_file_contents += [
            f"a a {float(self.op_point.alpha)}",
            f"b b {float(self.op_point.beta)}",
            f"r r {float(p_bar)}",
            f"p p {float(q_bar)}",
            f"y y {float(r_bar)}"
        ]

        # Set control surface deflections
        run_file_contents += [
            f"d1 d1 1"
        ]

        return run_file_contents

    def write_avl(self,
                  filepath: Union[Path, str] = None,
                  ) -> None:
        """
        Writes a .avl file corresponding to this airplane to a filepath.

        For use with the AVL vortex-lattice-method aerodynamics analysis tool by Mark Drela at MIT.
        AVL is available here: https://web.mit.edu/drela/Public/web/avl/

        Args:
            filepath: filepath (including the filename and .avl extension) [string]
                If None, this function returns the .avl file as a string.

        Returns: None

        """

        def clean(s):
            """
            Removes leading and trailing whitespace from each line of a multi-line string.
            """
            return "\n".join([line.strip() for line in s.split("\n")])

        airplane = self.airplane

        avl_file = ""

        airplane_options = self.get_options(airplane)

        avl_file += clean(f"""\
        {airplane.name}
        #Mach
        0        ! AeroSandbox note: This is overwritten later to match the current OperatingPoint Mach during the AVL run.
        #IYsym   IZsym   Zsym
         0       {1 if self.ground_effect else 0}   {self.ground_effect_height}
        #Sref    Cref    Bref
        {airplane.s_ref} {airplane.c_ref} {airplane.b_ref}
        #Xref    Yref    Zref
        {self.xyz_ref[0]} {self.xyz_ref[1]} {self.xyz_ref[2]}
        # CDp
        {airplane_options["profile_drag_coefficient"]}
        """)

        control_surface_counter = 0
        airfoil_counter = 0

        for wing in airplane.wings:

            wing_options = self.get_options(wing)

            spacing_line = f"{wing_options['chordwise_resolution']}   {self.AVL_spacing_parameters[wing_options['chordwise_spacing']]}"
            if wing_options["wing_level_spanwise_spacing"]:
                spacing_line += f"   {wing_options['spanwise_resolution']}   {self.AVL_spacing_parameters[wing_options['spanwise_spacing']]}"

            avl_file += clean(f"""\
            #{"=" * 79}
            SURFACE
            {wing.name}
            #Nchordwise  Cspace  [Nspanwise   Sspace]
            {spacing_line}
            
            """)

            if wing_options["component"] is not None:
                avl_file += clean(f"""\
                COMPONENT
                {wing_options['component']}
                    
                """)

            if wing.symmetric:
                avl_file += clean(f"""\
                YDUPLICATE
                0
                    
                """)

            if wing_options["no_wake"]:
                avl_file += clean(f"""\
                NOWAKE
                
                """)

            if wing_options["no_alpha_beta"]:
                avl_file += clean(f"""\
                NOALBE
                
                """)

            if wing_options["no_load"]:
                avl_file += clean(f"""\
                NOLOAD
                
                """)

            polar = wing_options["drag_polar"]
            avl_file += clean(f"""\
            CDCL
            #CL1  CD1  CL2  CD2  CL3  CD3
            {polar["CL1"]} {polar["CD1"]} {polar["CL2"]} {polar["CD2"]} {polar["CL3"]} {polar["CD3"]}
            
            """)

            ### Build up a buffer of the control surface strings to write to each section
            control_surface_commands: List[List[str]] = [
                []
                for _ in wing.xsecs
            ]
            for i, xsec in enumerate(wing.xsecs[:-1]):
                for surf in xsec.control_surfaces:
                    xhinge = surf.hinge_point if surf.trailing_edge else -surf.hinge_point
                    sign_dup = 1 if surf.symmetric else -1

                    command = clean(f"""\
                        CONTROL
                        #name, gain, Xhinge, XYZhvec, SgnDup
                        {surf.name} 1 {xhinge:.8g} 0 0 0 {sign_dup}
                        """)

                    control_surface_commands[i].append(command)
                    control_surface_commands[i + 1].append(command)

            ### Write the commands for each wing section
            for i, xsec in enumerate(wing.xsecs):

                xsec_options = self.get_options(xsec)

                xsec_def_line = f"{xsec.xyz_le[0]:.8g} {xsec.xyz_le[1]:.8g} {xsec.xyz_le[2]:.8g} {xsec.chord:.8g} {xsec.twist:.8g}"
                if not wing_options["wing_level_spanwise_spacing"]:
                    xsec_def_line += f"   {xsec_options['spanwise_resolution']}   {self.AVL_spacing_parameters[xsec_options['spanwise_spacing']]}"



                if xsec_options["cl_alpha_factor"] is None:
                    claf_line = f"{1 + 0.77 * xsec.airfoil.max_thickness()}  # Computed using rule from avl_doc.txt"
                else:
                    claf_line = f"{xsec_options['cl_alpha_factor']}"

                af_filepath = Path(str(filepath) + f".af{airfoil_counter}")
                airfoil_counter += 1
                xsec.airfoil.repanel(50).write_dat(filepath=af_filepath, include_name=True)

                avl_file += clean(f"""\
                #{"-" * 50}
                SECTION
                #Xle    Yle    Zle     Chord   Ainc  [Nspanwise   Sspace]
                {xsec_def_line}
                
                AFIL
                {af_filepath}
                
                CLAF
                {claf_line}
                
                """)

                polar = xsec_options["drag_polar"]
                avl_file += clean(f"""\
                CDCL
                #CL1  CD1  CL2  CD2  CL3  CD3
                {polar["CL1"]} {polar["CD1"]} {polar["CL2"]} {polar["CD2"]} {polar["CL3"]} {polar["CD3"]}
                
                """)

                for control_surface_command in control_surface_commands[i]:
                    avl_file += control_surface_command

        filepath = Path(filepath)
        for i, fuse in enumerate(airplane.fuselages):
            fuse_filepath = Path(str(filepath) + f".fuse{i}")
            self.write_avl_bfile(
                fuselage=fuse,
                filepath=fuse_filepath
            )
            fuse_options = self.get_options(fuse)

            avl_file += clean(f"""\
            #{"=" * 50}
            BODY
            {fuse.name}
            {fuse_options['panel_resolution']} {self.AVL_spacing_parameters[fuse_options['panel_spacing']]}
            
            BFIL
            {fuse_filepath}
            
            TRANSLATE
            0 {np.mean([x.xyz_c[1] for x in fuse.xsecs]):.8g} 0
            
            """)

        if filepath is not None:
            with open(filepath, "w+") as f:
                f.write(avl_file)

    @staticmethod
    def write_avl_bfile(fuselage,
                        filepath: Union[Path, str] = None,
                        include_name: bool = True,
                        ) -> str:
        """
        Writes an AVL-compatible BFILE corresponding to this fuselage to a filepath.

        For use with the AVL vortex-lattice-method aerodynamics analysis tool by Mark Drela at MIT.
        AVL is available here: https://web.mit.edu/drela/Public/web/avl/

        Args:
            filepath: filepath (including the filename and .avl extension) [string]
                If None, this function returns the would-be file contents as a string.

            include_name: Should the name of the fuselage be included in the .dat file? (This should be True for use with AVL.)

        Returns:

        """
        filepath = Path(filepath)

        contents = []

        if include_name:
            contents += [fuselage.name]

        contents += [
                        f"{xyz_c[0]:.8g} {xyz_c[2] + r:.8g}"
                        for xyz_c, r in zip(
                [xsec.xyz_c for xsec in fuselage.xsecs][::-1],
                [xsec.equivalent_radius(preserve="area") for xsec in fuselage.xsecs][::-1]
            )
                    ] + [
                        f"{xyz_c[0]:.8g} {xyz_c[2] - r:.8g}"
                        for xyz_c, r in zip(
                [xsec.xyz_c for xsec in fuselage.xsecs][1:],
                [xsec.equivalent_radius(preserve="area") for xsec in fuselage.xsecs][1:]
            )
                    ]

        string = "\n".join(contents)

        if filepath is not None:
            with open(filepath, "w+") as f:
                f.write(string)

        return string

    @staticmethod
    def parse_unformatted_data_output(
            s: str,
            data_identifier: str = " = ",
            cast_outputs_to_float: bool = True,
            overwrite: bool = None
    ) -> Dict[str, float]:
        """
        Parses a (multiline) string of unformatted data into a nice and tidy dictionary.

        The expected input string looks like what you might get as an output from AVL (or many other Drela codes),
        which may list data in ragged order.

        An example input `s` that you might want to parse could look like the following:

        ```
         Standard axis orientation,  X fwd, Z down

         Run case:  -unnamed-

          Alpha =   0.43348     pb/2V =  -0.00000     p'b/2V =  -0.00000
          Beta  =   0.00000     qc/2V =   0.00000
          Mach  =     0.003     rb/2V =  -0.00000     r'b/2V =  -0.00000

          CXtot =  -0.02147     Cltot =   0.00000     Cl'tot =   0.00000
          CYtot =   0.00000     Cmtot =   0.28149
          CZtot =  -1.01474     Cntot =  -0.00000     Cn'tot =  -0.00000

          CLtot =   1.01454
          CDtot =   0.02915
          CDvis =   0.00000     CDind = 0.0291513
          CLff  =   1.00050     CDff  = 0.0297201    | Trefftz
          CYff  =   0.00000         e =    0.9649    | Plane
        ```

        Here, this function will go through this string and extract each key-value pair, as denoted by the data
        identifier (by default, " = "). It will pull the next whole word without spaces to the left as the key,
        and it will pull the next whole word without spaces to the right as the value. Together, these will be
        returned as a Dict.

        So, the output for the input above would be:
        {
            'Alpha' : 0.43348,
            'pb/2V' : -0.00000,
            'p'b/2V' : -0.00000,
            'Beta' : 0.00000,
            # and so on...
        }

        Args:

            s: The input string to identify. Can be multiline.

            data_identifier: The triggering substring for a new key-value pair. By default, it's " = ",
            which is convention in many output files from Mark Drela's codes. Be careful if you decide to change this
            to "=", as you could pick up on heading separators ('=======') in Markdown-like files.

            cast_outputs_to_float: If this boolean flag is set true, the values of the key-value pairs are cast to
            floating-point numbers before returning (as opposed to the default type, string). If a value can't be
            cast, a NaN is returned (guaranteeing that you can do floating-point math with the outputs in downstream
            applications.)

            overwrite: Determines the behavior if you find a key that's already in the dictionary.

                * By default, value is None. In this case, an error is raised.

                * If you set it to True, the new value will overwrite the old one. Thus, your dictionary will have
                the last matching value from the string.

                * If you set it to False, the new value will be discarded. Thus, your dictionary will have the first
                matching value from the string.

        Returns: A dictionary of key-value pairs, corresponding to the unformatted data in the input string.

            Keys are strings, values are floats if `cast_outputs_to_float` is True, otherwise also strings.

        """

        items = {}

        index = s.find(data_identifier)

        while index != -1:  # While there are still data identifiers:

            key = ""  # start with a blank key, which we will build up as we read

            i = index - 1  # Starting from the left of the identifier
            while s[i] == " " and i >= 0:
                # First, skip any blanks
                i -= 1
            while s[i] != " " and s[i] != "\n" and i >= 0:
                # Then, read the key in backwards order until you get to a blank or newline
                key = s[i] + key
                i -= 1

            value = ""  # start with a blank value, which we will build up as we read

            i = index + len(data_identifier)  # Starting from the right of the identifier
            while s[i] == " " and i <= len(s):
                # First, skip any blanks
                i += 1
            while s[i] != " " and s[i] != "\n" and i <= len(s):
                # Then, read the key in forward order until you get to a blank or newline
                value += s[i]
                i += 1

            if cast_outputs_to_float:
                try:  # Try to convert the value into a float. If you can't, return a NaN
                    value = float(value)
                except Exception:
                    value = np.nan

            if key in items.keys():  # If you already have this key
                if overwrite is None:  # If the `overwrite` parameter wasn't explicitly defined True/False, raise an error
                    raise ValueError(
                        f"Key \"{key}\" is being overwritten, and no behavior has been specified here (Default behavior is to error).\n"
                        f"Check that the output file doesn't have a duplicate here.\n"
                        f"Alternatively, set the `overwrite` parameter of this function to True or False (rather than the default None).",
                    )
                else:
                    if overwrite:
                        items[key] = value  # Assign (and overwrite) the key-value pair to the output we're writing
                    else:
                        pass
            else:
                items[key] = value  # Assign the key-value pair to the output we're writing

            s = s[index + len(data_identifier):]  # Trim the string by starting to read from the next point.
            index = s.find(data_identifier)

        return items


if __name__ == '__main__':

    ### Import Vanilla Airplane
    import aerosandbox as asb

    from aerosandbox.aerodynamics.aero_3D.test_aero_3D.geometries.vanilla import airplane as vanilla

    vanilla.analysis_specific_options[AVL] = dict(
        profile_drag_coefficient=0.1
    )
    vanilla.wings[0].xsecs[0].control_surfaces.append(
        ControlSurface(
            name="Flap",
            trailing_edge=True,
            hinge_point=0.75,
            symmetric=True,
            deflection=10
        )
    )
    ### Do the AVL run
    avl = AVL(
        airplane=vanilla,
        op_point=OperatingPoint(
            atmosphere=asb.Atmosphere(altitude=0),
            velocity=1,
            alpha=0.433476,
            beta=0,
            p=0,
            q=0,
            r=0,
        ),
        working_directory=str(Path.home() / "Downloads" / "test"),
        verbose=True
    )

    res = avl.run()

    for k, v in res.items():
        print(f"{str(k).rjust(10)} : {v}")



import aerosandbox.numpy as np
from aerosandbox.geometry.airplane import Airplane
from aerosandbox.geometry.wing import Wing, WingXSec
from aerosandbox.geometry.fuselage import Fuselage, FuselageXSec
from aerosandbox.geometry.propulsor import Propulsor
from aerosandbox.geometry.airfoil.airfoil import Airfoil
from aerosandbox.geometry.airfoil.kulfan_airfoil import KulfanAirfoil
from textwrap import indent, dedent

# noindent

_openvsp_version = "3.36.0"


def wrap_script(
        script: str,
        set_geom_draw_type_to_shade: bool = True,
) -> str:
    """
    Wraps the internal parts of a VSPScript file with a main() function.

    Example:
        >>> print(wrap_script("hello\nworld"))
        // This file was automatically generated by AeroSandbox 4.2.0
        // using syntax tested on OpenVSP 3.36.0.

        void main()
        {
            hello
            world
        }

    Args:
        script: The script to wrap.

    Returns: The script, wrapped with a main() function.

    """
    script = script + """\
    
//==== Check For API Errors ====//
while ( GetNumTotalErrors() > 0 )
{
    ErrorObj err = PopLastError();
    Print( err.GetErrorString() );
}
"""

    if set_geom_draw_type_to_shade:
        script = script + """\

{
    array<string> @geomids = FindGeoms();
    
    for (uint i = 0; i < geomids.length(); i++)
    {
        SetGeomDrawType( geomids[i], GEOM_DRAW_SHADE );
    }
}
        
"""

    import aerosandbox as asb
    return f"""\
// This *.vspscript file was automatically generated by AeroSandbox {asb.__version__} 
// using syntax tested on OpenVSP {_openvsp_version}.
// To run this script, open OpenVSP and go to File -> Run Script...

void main()
{{
{indent(script, "    ")}
}}
"""


if __name__ == '__main__':

    print(wrap_script("hello\nworld"))  # Expected:
    # // This file was automatically generated by AeroSandbox 4.2.0
    # // using syntax tested on OpenVSP 3.36.0.
    #
    # void main()
    # {
    #     hello
    #     world
    #
    #
    #     //==== Check For API Errors ====//
    #     while ( GetNumTotalErrors() > 0 )
    #     {
    #         ErrorObj err = PopLastError();
    #         Print( err.GetErrorString() );
    #     }
    #



from .vortex_lattice_method import VortexLatticeMethod
from .lifting_line import LiftingLine
from .nonlinear_lifting_line import NonlinearLiftingLine
from .aero_buildup import AeroBuildup
from .avl import AVL



import aerosandbox.numpy as np
import casadi as _cas
from typing import Union, Callable, Tuple, Optional, Dict, Any, List
from scipy import integrate


def quad(
        func: Union[Callable, _cas.MX],
        a: float,
        b: float,
        full_output: bool = False,
        variable_of_integration: _cas.MX = None,
) -> Union[
    Tuple[float, float],
    Tuple[float, float, dict]
]:
    if np.is_casadi_type(func):

        all_vars = _cas.symvar(func)  # All variables found in the expression graph

        if variable_of_integration is None:
            if not len(all_vars) == 1:
                raise ValueError(
                    f"`func` must be a function of one variable, or you must specify the `variable_of_integration`.\n"
                    f"Currently, it is a function of: {all_vars}")
            variable_of_integration = all_vars[0]

        parameters = [
            var for var in all_vars
            if not _cas.is_equal(var, variable_of_integration)
        ]

        integrator = _cas.integrator(
            'integrator',
            'cvodes',
            {
                'x'  : _cas.MX.sym('dummy_variable'),
                'p'  : _cas.vertcat(*parameters),
                't'  : variable_of_integration,
                'ode': func,
            },
            a,  # t0
            b,  # tf
            {  # Options
                'abstol': 1e-8,
                'reltol': 1e-6,
            },
        )
        res = integrator(
            x0=0,
            p=_cas.vertcat(*parameters),
        )
        tol = 1e-8

        if full_output:
            return res['xf'], tol, res
        else:
            return res['xf'], tol

    else:
        return integrate.quad(
            func=func,
            a=a,
            b=b,
            full_output=full_output,
        )


def solve_ivp(
        fun: Union[Callable, _cas.MX],
        t_span: Tuple[float, float],
        y0: Union[np.ndarray, _cas.MX],
        method: str = 'RK45',
        t_eval: Union[np.ndarray, _cas.MX] = None,
        dense_output: bool = False,
        events: Union[Callable, List[Callable]] = None,
        vectorized: bool = False,
        args: Optional[Tuple] = None,
        t_variable: _cas.MX = None,
        y_variables: Union[_cas.MX, Tuple[_cas.MX]] = None,
        **options
):

    # Determine which backend to use
    if np.is_casadi_type(fun, recursive=False):
        backend = "casadi_expr"
    else:
        try:
            f = np.array(fun(t_span[0], y0))
            if np.is_casadi_type(f):
                backend = "casadi_func"
            else:
                try:
                    np.asanyarray(f)
                except ValueError:
                    raise ValueError("If `fun` is not a Callable, it must be a CasADi expression.")
                backend = "numpy_func"
        except TypeError:
            raise TypeError("If `fun` is not a Callable, it must be a CasADi expression.")

    # Do some checks
    if backend == "casadi_func" or backend == "numpy_func":
        if t_variable is not None:
            raise ValueError("If `fun` is a Callable, `t_variable` must be None (as it's implied).")
        if y_variables is not None:
            raise ValueError("If `fun` is a Callable, `y_variables` must be None (as they're implied).")

    if backend == "casadi_expr":
        if t_variable is None:
            raise ValueError(
                "If `fun` is a CasADi expression, `t_variable` must be specified (and the y_variables are inferred).")

        all_vars = _cas.symvar(fun)  # All variables found in the expression graph

        # Determine y_variables by selecting all variables that are not t_variable
        if y_variables is None:
            y_variables = np.array([
                var for var in all_vars
                if not _cas.is_equal(var, t_variable)
            ])

    if backend == "numpy_func":
        return integrate.solve_ivp(
            fun=fun,
            t_span=t_span,
            y0=y0,
            method=method,
            t_eval=t_eval,
            dense_output=dense_output,
            events=events,
            vectorized=vectorized,
            args=args,
            **options
        )
    elif backend == "casadi_func" or backend == "casadi_expr":

        # Exception on non-implemented options
        if dense_output:
            raise NotImplementedError("dense_output is not yet implemented for CasADi functions.")
        if events is not None:
            raise NotImplementedError("Events are not yet implemented for CasADi functions.")
        if args:
            raise NotImplementedError("args are not yet implemented for CasADi functions.")

        if not np.is_casadi_type(y0, recursive=False):
            y0 = _cas.vertcat(*y0)

        if backend == "casadi_func":

            t_variable = _cas.MX.sym("t")
            y_variables = _cas.MX.sym("y", y0.shape[0], y0.shape[1])
            fun = np.array(fun(t_variable, y_variables))

        """
        At this point:
        * `fun` is a CasADi expression (cas.MX)
        * `t_variable` is a CasADi variable (cas.MX)
        * `y_variables` is a CasADi variable (cas.MX), possibly a vector of variables
        """

        t0 = t_span[0]
        tf = t_span[1]

        # sim_time = t0 + (tf - t0) * t_variable
        ode = _cas.substitute(
            fun,
            t_variable,  # from normalized time
            # (t_variable - t0) / (tf - t0), # to real time
            t0 + (tf - t0) * t_variable,  # to real time
        ) * (tf - t0)

        # Find parameters by finding all variables in the expression graph that are not t_variable or y_variables
        all_vars = _cas.symvar(ode)  # All variables found in the expression graph

        def variable_is_t_or_y(var):
            return (
                    _cas.is_equal(var, t_variable) or
                    _cas.is_equal(var, y_variables) or
                    any([_cas.is_equal(var, y_variables[i]) for i in range(np.prod(y_variables.shape))])
            )

        parameters = _cas.vertcat(*[
            var for var in all_vars
            if not variable_is_t_or_y(var)
        ])

        simtime_eval = np.linspace(0, 1, 100)

        # Define the integrator
        integrator = _cas.integrator(
            'integrator',
            'cvodes',
            # 'idas',
            {
                'x'   : y_variables,
                'p'   : parameters,
                't'   : t_variable,
                'ode' : ode,
                'quad': 1,
            },
            0,
            simtime_eval,
            {  # Options
                'abstol': 1e-8,
                'reltol': 1e-6,
            },
        )
        res = integrator(
            x0=y0,
            p=parameters,
        )

        return integrate._ivp.ivp.OdeResult(
            t=t0 + (tf - t0) * res["qf"],
            y=res['xf'],
            t_events=None,
            y_events=None,
            nfev=0,
            njev=0,
            nlu=0,
            status=0,
            message="",
            success=True,
            sol=None,
        )

    else:
        raise ValueError(f"Invalid backend: {backend}")


if __name__ == '__main__':
    # t = cas.MX.sym("t")
    # print(
    #     quad(
    #         func=t ** 2,
    #         a=0,
    #         b=1,
    #     )
    # )

    def lotkavolterra_func(t, z):
        a, b, c, d = 1.5, 1, 3, 1
        z = _cas.MX(z)
        x = z[0]
        y = z[1]
        return [a * x - b * x * y, -c * y + d * x * y]


    t_eval = np.linspace(0, 15, 3000)
    tf = _cas.MX.sym("tf")
    # t_eval = np.linspace(0, tf, 100)

    sol = solve_ivp(
        lotkavolterra_func,
        t_span=(t_eval[0], t_eval[-1]),
        # t_eval=t_eval,
        y0=[10, 5],
    )

    z = sol.y
    import matplotlib.pyplot as plt

    plt.plot(
        _cas.evalf(_cas.substitute(sol.t.T, tf, 15)),
        _cas.evalf(_cas.substitute(sol.y.T, tf, 15)),
    )
    plt.xlabel('t')
    plt.legend(['x', 'y'], shadow=True)
    plt.title('Lotka-Volterra System')
    plt.show()

    t = _cas.MX.sym("t")
    m = _cas.MX.sym("m")
    n = _cas.MX.sym("n")
    a, b, c, d = 1.5, 1, 3, 1
    lotkavolterra_expr = np.array([
        a * m - b * m * n,
        -c * n + d * m * n,
    ])

    sol = solve_ivp(
        lotkavolterra_expr,
        t_span=(t_eval[0], t_eval[-1]),
        t_eval=t_eval,
        y0=[10, 5],
        t_variable=t,
        # y_variables=[m, n],
    )

    plt.plot(
        _cas.evalf(_cas.substitute(sol.t.T, tf, 15)),
        _cas.evalf(_cas.substitute(sol.y.T, tf, 15)),
    )
    plt.xlabel('t')
    plt.legend(['x', 'y'], shadow=True)
    plt.title('Lotka-Volterra System')
    plt.show()



from aerosandbox.geometry.common import reflect_over_XZ_plane
import aerosandbox.numpy as np
import pytest
import casadi as cas

vec = np.arange(3)
square = np.arange(9).reshape((3, 3))
rectangular_tall = np.arange(12).reshape((4, 3))
rectangular_wide = np.arange(12).reshape((3, 4))


def test_np_vector():
    assert np.all(
        reflect_over_XZ_plane(vec) ==
        np.array([0, -1, 2])
    )


def test_cas_vector():
    output = reflect_over_XZ_plane(cas.DM(vec))
    assert isinstance(output, cas.DM)
    assert np.all(
        output ==
        np.array([0, -1, 2])
    )


def test_np_vector_2D_wide():
    assert np.all(
        reflect_over_XZ_plane(np.expand_dims(vec, 0)) ==
        np.array([0, -1, 2])
    )


def test_np_vector_2D_tall():
    with pytest.raises(ValueError):
        reflect_over_XZ_plane(np.expand_dims(vec, 1))


def test_np_square():
    assert np.all(
        reflect_over_XZ_plane(square) ==
        np.array([
            [0, -1, 2],
            [3, -4, 5],
            [6, -7, 8]
        ])
    )


def test_cas_square():
    output = reflect_over_XZ_plane(cas.DM(square))
    assert isinstance(output, cas.DM)
    assert np.all(
        output ==
        np.array([
            [0, -1, 2],
            [3, -4, 5],
            [6, -7, 8]
        ])
    )


def test_np_rectangular_tall():
    assert np.all(
        reflect_over_XZ_plane(rectangular_tall) ==
        np.array([
            [0, -1, 2],
            [3, -4, 5],
            [6, -7, 8],
            [9, -10, 11],
        ])
    )


def test_cas_rectangular_tall():
    output = reflect_over_XZ_plane(cas.DM(rectangular_tall))
    assert isinstance(output, cas.DM)
    assert np.all(
        output ==
        np.array([
            [0, -1, 2],
            [3, -4, 5],
            [6, -7, 8],
            [9, -10, 11],
        ])
    )


def test_np_rectangular_wide():
    with pytest.raises(ValueError):
        reflect_over_XZ_plane(rectangular_wide)


def test_cas_rectangular_wide():
    with pytest.raises(ValueError):
        reflect_over_XZ_plane(cas.DM(rectangular_wide))


def test_np_3D():
    with pytest.raises(ValueError):
        reflect_over_XZ_plane(np.arange(2 * 3 * 4).reshape((2, 3, 4)))


if __name__ == '__main__':
    pytest.main()



import inspect
from typing import List, Union, Tuple, Optional, Set, Any
from pathlib import Path
from aerosandbox.tools.string_formatting import has_balanced_parentheses
import numpy as np


def get_caller_source_location(
        stacklevel: int = 1,
) -> (Path, int, str):
    """
    Gets the file location where this function itself (`get_caller_source_location()`) is called.

    This is not usually useful by itself. However, with the use of the `stacklevel` argument, you can get the call
    location at any point arbitrarily high up in the call stack from this function.

    This potentially lets you determine the file location where any Python object was declared.

    Examples:

        Consider the file below (and assume we somehow have this function in scope):

        my_file.py:
        >>> def my_func():
        >>>     print(
        >>>         get_caller_source_location(stacklevel=2)
        >>>     )
        >>>
        >>> if_you_can_see_this_it_works = my_func()

        This will print out the following:
        (/path/to/my_file.py, 5, "if_you_can_see_this_it_works = my_func()\n")

    Args:

        stacklevel: Choose the level of the stack that you want to retrieve source code at. Higher integers will get
        you higher (i.e., more end-user-facing) in the stack. Same behaviour as the `stacklevel` argument in
        warnings.warn().

    Returns: A tuple of:
        (filename, lineno, code_context)

        * `filename`: a Path object (see `pathlib.Path` from the standard Python library) of the file where this function was called.

        * `lineno`: the line number in the file where this function was called.

        * `code_context`: the immediate line of code where this function was called. A string. Note that, in the case of
        multiline statements, this may not be a complete Python expression. Includes the trailing newline character ("\n") at the end.

    """
    ### Go up `stacklevel` frames from the current one to get to the caller frame.
    frame = inspect.currentframe()
    for _ in range(stacklevel):
        frame = frame.f_back

    ### Extract the frame info (an `inspect.Traceback` type) from the caller frame
    frame_info: inspect.Traceback = inspect.getframeinfo(frame)

    filename = Path(frame_info.filename)
    lineno = frame_info.lineno
    if frame_info.code_context is not None:
        code_context = "".join(frame_info.code_context)
    else:
        code_context = ""

    return filename, lineno, code_context


def get_source_code_from_location(
        filename: Union[Path, str],
        lineno: int,
        code_context: str = None,
        strip_lines: bool = False
) -> str:
    """
    Gets the source code of the single statement that begins at the file location specified.

    File location must, at a minimum, contain the filename and the line number. Optionally, you can also provide `code_context`.

     These should have the format:

        * `filename`: a Path object (see `pathlib.Path` from the standard Python library) of the file where this function was called.

        * `lineno`: the line number in the file where this function was called.

    Optionally, you can also provide `code_context`, which has the format:

        * `code_context`: the immediate line of code where this function was called. A string. Note that, in the case of
        multiline statements, this may not be a complete Python expression.

    You can get source code from further up the call stack by using the `stacklevel` argument.

    Args:

        filename: a Path object (see `pathlib.Path` from the standard Python library) of the file where this function
        was called. Alternatively, a string containing a filename.

        lineno: the line number in the file where this function was called. An integer. Should refer to the first
        line of a string in question.

        code_context: Optional. Should be a string containing the immediate line of code at this location. If
        provided, allows short-circuiting (bypassing file I/O) if the line is a complete expression.

        strip_lines: A boolean flag about whether or not to strip leading and trailing whitespace off each line of a
        multi-line function call. See the built-in string method `str.strip()` for behaviour.

    Returns: The source code of the call, as a string. Might be a multi-line string (i.e., contains '\n' characters)
    if the call is multi-line. Almost certainly (but not guaranteed due to edge cases) to be a complete Python expression.
    """

    ### If Python's auto-extracted "code context" is a compete statement, then you're done here.
    if code_context is not None:
        if has_balanced_parentheses(code_context):
            return code_context

    ### Initialize the caller source lines, which is a list of strings that contain the source for the call.
    source_lines: List[str] = []

    ### Read the source lines of code surrounding the call
    try:
        with open(filename, "r") as f:  # Read the file containing the call
            for _ in range(lineno - 1):  # Skip the first N lines of code, until you get to the call
                f.readline()  # Unfortunately there's no way around this, since you need to find the "\n" encodings in the file

            parenthesis_level = 0  # Track the number of "(" and ")" characters, so you know when the function call is complete

            def add_line() -> None:
                """
                Adds the subsequent line to the caller source lines (`caller_source_lines`). In-place.
                """
                line = f.readline()

                if strip_lines:
                    line = line.strip()

                nonlocal parenthesis_level  # TODO add "\" support
                for char in line:
                    if char == "(":
                        parenthesis_level += 1
                    elif char == ")":
                        parenthesis_level -= 1
                source_lines.append(line)

            ### Get the first line, which is always part of the function call, and includes the opening parenthesis
            add_line()

            ### Do subsequent lines
            while parenthesis_level > 0:
                add_line()
    except OSError as e:
        raise FileNotFoundError(
            "\n".join([
                "Couldn't retrieve source code at this stack level, because the source code file couldn't be opened for some reason.",
                "One common possible reason is that you're referring to an IPython console with a multi-line statement."
            ])
        )

    source = "".join(source_lines)

    return source


def get_caller_source_code(
        stacklevel: int = 1,
        strip_lines: bool = False
) -> str:
    """
    Gets the source code of wherever this function is called.

    You can get source code from further up the call stack by using the `stacklevel` argument.

    Args:

        stacklevel: Choose the level of the stack that you want to retrieve source code at. Higher integers will get
        you higher (i.e., more end-user-facing) in the stack. Same behaviour as the `stacklevel` argument in
        warnings.warn().

        strip_lines: A boolean flag about whether or not to strip leading and trailing whitespace off each line of a
        multi-line function call. See the built-in string method `str.strip()` for behaviour.

    Returns: The source code of the call, as a string. Might be a multi-line string (i.e., contains '\n' characters)
    if the call is multi-line. Almost certainly (but not guaranteed due to edge cases) to be a complete Python expression.
    """

    filename, lineno, code_context = get_caller_source_location(
        stacklevel=stacklevel + 1
    )

    return get_source_code_from_location(
        filename=filename,
        lineno=lineno,
        code_context=code_context,
        strip_lines=strip_lines
    )


def get_function_argument_names_from_source_code(source_code: str) -> List[str]:
    """
    Gets the names of the function arguments found in a particular line of source code.

    Specifically, it retrieves the names of the arguments in the first function call found in the source code string.

    If the source code line is an assignment statement, only the right-hand-side of the line is analyzed.

    Also, removes all line breaks ('\n').

    Examples function inputs and outputs:

        "f(a, b)"                -> ['a', 'b']
        "f(a,b)"                 -> ['a', 'b']
        "f(\na,\nb)"             -> ['a', 'b']
        "g = f(a, b)"            -> ['a', 'b']
        "g.h = f(a, b)"          -> ['a', 'b']
        "g.h() = f(a, b)"        -> ['a', 'b']
        "g.h(i=j) = f(a, b)"     -> ['a', 'b']
        "f(a, b) + g(h)"         -> ['a', 'b']
        "f(a: int, b: MyType())" -> ['a', 'b']
        "f(a, b).g(c, d)"        -> ['a', 'b']
        "f(a(b), c)"             -> ['a(b)', 'c']
        "f(a(b,c), d)"           -> ['a(b,c)', 'd']
        "f({a:b}, c)"            -> ['{a:b}', 'c']
        "f(a[b], c)"             -> ['a[b]', 'c']
        "f({a:b, c:d}, e)"       -> ['{a:b, c:d}', 'e']
        "f({a:b,\nc:d}, e)"      -> ['{a:b,c:d}', 'e']
        "f(dict(a=b,c=d), e)"    -> ['dict(a=b,c=d)', 'e']
        "f(a=1, b=2)"            -> ['a=1', 'b=2']
        "f()"                    -> ['']
        "f(a, [i for i in l])"   -> ['a', '[i for i in l]'],
        "f(incomplete, "         -> raises ValueError
        "3 + 5"                  -> raises ValueError
        ""                       -> raises ValueError

    Args:
        source_code: A line of Python source code that includes a function call. Can be a multi-line piece of source code (e.g., includes '\n').

    Returns: A list of strings containing all the function arguments. If keyword arguments are found, includes both the key and the value, as-written.

    """

    assignment_equals_index = 0

    parenthesis_level = 0
    for i, char in enumerate(source_code):
        if char == "(":
            parenthesis_level += 1
        elif char == ")":
            parenthesis_level -= 1
        elif char == "=" and parenthesis_level == 0:
            assignment_equals_index = i + 1
            break

    source_code_rhs = source_code[assignment_equals_index:]

    source_code_rhs = source_code_rhs.replace("\n", "")

    parenthesis_level = 0
    braces_level = 0
    for i, char in enumerate(source_code_rhs):
        if char == "(":
            parenthesis_level += 1
            break

    if parenthesis_level == 0:
        raise ValueError("No function call was found in the source code provided!")

    arg_names: List[str] = []
    current_arg = ""
    in_type_hinting_block = False

    while parenthesis_level != 0:
        i += 1
        if i >= len(source_code_rhs):
            raise ValueError("Couldn't match all parentheses, so this doesn't look like valid code!")
        char = source_code_rhs[i]

        if char == "(":
            parenthesis_level += 1
        elif char == ")":
            parenthesis_level -= 1
        elif char == "{":
            braces_level += 1
        elif char == "}":
            braces_level -= 1

        if char == "," and parenthesis_level == 1 and braces_level == 0:
            arg_names.append(current_arg)
            current_arg = ""
            in_type_hinting_block = False
        elif char == ":" and parenthesis_level == 1 and braces_level == 0:
            in_type_hinting_block = True
        elif parenthesis_level >= 1 and not in_type_hinting_block:
            current_arg += char

    arg_names.append(current_arg.strip())

    def clean(s: str) -> str:
        return s.strip()

    arg_names = [
        clean(arg) for arg in arg_names
    ]

    return arg_names


def codegen(
        x: Any,
        indent_str: str = "    ",
        _required_imports: Optional[Set[str]] = None,
        _recursion_depth: int = 0,
) -> Tuple[str, Set[str]]:
    """
    Attempts to generate a string of Python code that, when evaluated, would produce the same value as the input.
    Also generates the required imports for the code to run.

    In other words, in general, the following should evaluate True:

    >>> code, imports = codegen(x)
    >>> for import_str in imports:
    >>>     exec(import_str)
    >>> eval(code) == x  # Should evaluate True

    Not guaranteed to work for all inputs, but should work for most common cases.

    Args:

        x: The object to generate the code of.

        indent_str: The string to use for indentation. Defaults to four spaces.

        _required_imports: A set of strings containing the names of all required imports. This is an internal
        argument that should not be used by the user.

        _recursion_depth: The current recursion depth. This is an internal argument that should not be used by the user.

    Returns: A tuple containing:

        - The string of Python code that, when evaluated, would produce the same value as the input.

        - A set of strings that, when evaluated, would import all the required imports for the code to run.

    Examples:

        >>> codegen(5)
        ('5', set())

        >>> codegen([1, 2, 3])
        ('[1, 2, 3]', set())

        >>> codegen(np.array([1, 2, 3]))
        ('np.array([1, 2, 3])', {'import numpy as np'})

        >>> codegen(dict(my_int=4, my_array=np.array([1, 2, 3])))
        ('{
        	'my_int': 4,
        	'my_array': np.array([1, 2, 3]),
        }', {'import numpy as np'})

    """
    ### Set defaults
    if _required_imports is None:
        _required_imports = set()

    import_aliases = {
        "aerosandbox"      : "asb",
        "aerosandbox.numpy": "np",
        "numpy"            : "np",
    }

    indent = indent_str * _recursion_depth
    next_indent = indent_str * (_recursion_depth + 1)

    if isinstance(x, (
            bool, str,
            int, float, complex,
            range,
            type(None),
            bytes, bytearray, memoryview
    )):
        code = repr(x)

    elif isinstance(x, list):
        if len(x) == 0:
            code = "[]"
        else:
            lines = []
            lines.append("[")
            for xi in x:
                item_code, item_required_imports = codegen(xi, _recursion_depth=_recursion_depth + 1)

                _required_imports.update(item_required_imports)

                lines.append(next_indent + item_code + ",")
            lines.append(indent + "]")
            code = "\n".join(lines)

    elif isinstance(x, tuple):
        if len(x) == 0:
            code = "()"
        else:
            lines = []
            lines.append("(")
            for xi in x:
                item_code, item_required_imports = codegen(xi, _recursion_depth=_recursion_depth + 1)

                _required_imports.update(item_required_imports)

                lines.append(next_indent + item_code + ",")
            lines.append(indent + ")")
            code = "\n".join(lines)

    elif isinstance(x, (set, frozenset)):
        if len(x) == 0:
            code = "set()"
        else:
            lines = []
            lines.append("{")
            for xi in x:
                item_code, item_required_imports = codegen(xi, _recursion_depth=_recursion_depth + 1)

                _required_imports.update(item_required_imports)

                lines.append(next_indent + item_code + ",")
            lines.append(indent + "}")
            code = "\n".join(lines)

    elif isinstance(x, dict):
        if len(x) == 0:
            code = "{}"
        else:
            lines = []
            lines.append("{")
            for k, v in x.items():
                k_code, k_required_imports = codegen(k, _recursion_depth=_recursion_depth + 1)
                v_code, v_required_imports = codegen(v, _recursion_depth=_recursion_depth + 1)

                _required_imports.update(k_required_imports)
                _required_imports.update(v_required_imports)

                lines.append(next_indent + k_code + ": " + v_code + ",")
            lines.append(indent + "}")
            code = "\n".join(lines)

    elif isinstance(x, np.ndarray):
        _required_imports.add("import numpy as np")
        code = f"np.{repr(x)}"

    else:  # At this point, we assume it's a class instance, and could be from any package.

        module_name = x.__class__.__module__
        package_name = module_name.split(".")[0]

        if package_name == "builtins":
            pre_string = ""
        # elif package_name in import_aliases:
        #     pre_string = import_aliases[package_name] + "."
        else:
            _required_imports.add(
                f"from {module_name} import {x.__class__.__name__}"
            )

        lines = []
        lines.append(x.__class__.__name__ + "(")
        for arg_name in inspect.getfullargspec(x.__init__).args[1:]:
            if hasattr(x, arg_name):
                arg_value = getattr(x, arg_name)

                if inspect.ismethod(arg_value) or inspect.isfunction(arg_value):
                    continue

                arg_code, arg_required_imports = codegen(arg_value, _recursion_depth=_recursion_depth + 1)

                _required_imports.update(arg_required_imports)

                lines.append(next_indent + arg_name + "=" + arg_code + ",")
        lines.append(indent + ")")
        code = "\n".join(lines)

    return code, _required_imports
    #
    # if _recursion_depth == 0:
    #     if len(_required_imports) > 0:
    #         imports = "\n".join(sorted(_required_imports))
    #         return imports + "\n\n" + code
    #
    #     else:
    #         return code
    # else:
    #     return code, _required_imports


if __name__ == '__main__':
    def dashes():
        """A quick macro for drawing some dashes, to make the terminal output clearer to distinguish."""
        print("\n" + "-" * 50 + "\n")


    dashes()

    print("Caller location:\n", get_caller_source_location(stacklevel=1))

    dashes()

    print("Caller source code:\n", get_caller_source_code(stacklevel=1))

    dashes()


    def my_func():
        print(
            get_caller_source_code(
                stacklevel=2
            )
        )


    print("Caller source code of a function call:")

    if_you_can_see_this_it_works = my_func()

    dashes()

    print("Arguments of f(a, b):")

    print(
        get_function_argument_names_from_source_code("f(a, b)")
    )

    location = get_caller_source_location()

    dashes()

    print("Codegen test:")


    def pc(x):
        code, imports = codegen(x)
        print("\n".join(sorted(imports)))
        print(code + "\n" + "-" * 50)


    pc(1)
    pc([1, 2, 3])
    pc([1, 2, [3, 4, 5], 6])
    pc({"a": 1, "b": 2})
    pc(np.array([1, 2, 3]))
    pc(dict(myarray=np.array([1, 2, 3]), yourarray=np.arange(10)))



import time
from typing import Tuple, Any, Callable
import aerosandbox.numpy as np


class Timer(object):
    """
    A context manager for timing things. Use it like this:

    with Timer("My timer"):  # You can optionally give it a name
        # Do stuff

    Results are printed to stdout. You can access the runtime (in seconds) directly by instantiating the object:

    >>> t = Timer("My timer")
    >>> t.tic()
    >>> # Do stuff
    >>> print(t.toc())

    Nested timers are also supported. For example, this code:
    >>> with Timer("a"):
    >>>     with Timer("b"):
    >>>         with Timer("c"):
    >>>             f()

    prints the following console output:

        [a] Timing...
        	[b] Timing...
        		[c] Timing...
        		[c] Elapsed: 100 msec
        	[b] Elapsed: 100 msec
        [a] Elapsed: 100 msec

    """
    number_running: int = 0  # The number of Timers currently running

    def __init__(self,
                 name: str = None,
                 verbose: bool = True,
                 ):
        self.name: str = name
        self.verbose: bool = verbose
        self.runtime: float = np.nan

    def __repr__(self):
        return f"{self.__class__.__name__}: " + (
            "Running..."
            if np.isnan(self.runtime) else
            f"Finished, elapsed: ({self._format_time(self.runtime)})"
        )

    @staticmethod
    def _format_time(time_seconds):
        from aerosandbox.tools.string_formatting import eng_string
        from aerosandbox.tools import units as u

        if time_seconds < u.minute:
            return eng_string(time_seconds, unit="sec")
        elif time_seconds < u.hour:
            return eng_string(time_seconds / u.minute, unit="min")
        elif time_seconds < u.day:
            return eng_string(time_seconds / u.hour, unit="hour")
        elif time_seconds < u.year:
            return eng_string(time_seconds / u.day, unit="day")
        else:
            return eng_string(time_seconds / u.year, unit="year")

    def _print(self, s: str, number_running_mod: int = 0):
        if self.verbose:
            header = "\t" * (self.__class__.number_running - 1 + number_running_mod)
            if self.name:
                header += f"[{self.name}] "
            print(header + s)

    def tic(self):
        self.__class__.number_running += 1
        self._print("Timing...")
        self.t_start = time.perf_counter_ns()

    def __enter__(self):
        self.tic()

    def toc(self) -> float:
        self.t_end = time.perf_counter_ns()
        self.__class__.number_running -= 1
        self.runtime = (self.t_end - self.t_start) / 1e9
        self._print(
            f"Elapsed: {self._format_time(self.runtime)}",
            number_running_mod=1
        )
        return self.runtime

    def __exit__(self, type, value, traceback):
        self.toc()


def time_function(
        func: Callable,
        repeats: int = None,
        desired_runtime: float = None,
        runtime_reduction=np.min,
        warmup: bool = True,
) -> Tuple[float, Any]:
    """
    Runs a given callable and tells you how long it took to run, in seconds. Also returns the result of the function
        (if any), for good measure.

    Args:

        func: The function to run. Should take no arguments; use a lambda function or functools.partial if you need
            to pass arguments.

        repeats: The number of times to run the function. If None, runs until desired_runtime is met.

        desired_runtime: The desired runtime of the function, in seconds. If None, runs until repeats is met.

        runtime_reduction: A function that takes in a list of runtimes and returns a reduced value. For example,
            np.min will return the minimum runtime, np.mean will return the mean runtime, etc. Default is np.min.

        warmup: If True, runs the function once before starting the timer. This can be useful for functions with JIT-compilation,
            with internal imports, or caches.

    Returns: A Tuple of (time_taken, result).
        - time_taken is a float of the time taken to run the function, in seconds.
        - result is the result of the function, if any.

    """
    if (repeats is not None) and (desired_runtime is not None):
        raise ValueError("You can't specify both repeats and desired_runtime!")

    if warmup:
        func()

    def time_function_once():
        start_ns = time.perf_counter_ns()
        result = func()
        return (
            (time.perf_counter_ns() - start_ns) / 1e9,
            result
        )

    runtimes = []

    t, result = time_function_once()

    if t == 0:
        t = 1e-2
    else:
        runtimes.append(t)

    if (desired_runtime is not None) and (repeats is None):
        repeats = int(desired_runtime // t) - 1
        # print(f"Running {func.__name__} {repeats} times to get a desired runtime of {desired_runtime} seconds.")

    if repeats is None:
        repeats = 0

    for _ in range(repeats):
        t, _ = time_function_once()
        if t != 0:
            runtimes.append(t)

    if len(runtimes) == 0:
        runtimes = [0.]

    return (
        runtime_reduction(runtimes),
        result
    )


if __name__ == '__main__':

    def f():
        time.sleep(0.1)


    print(time_function(f, desired_runtime=1))

    with Timer("a") as a:
        with Timer("b") as b:
            with Timer("c") as c:
                f()

    t = Timer()
    t.tic()
    time.sleep(0.1)
    print(t.toc())



import warnings
from typing import Union, Iterable, List, Tuple

from tqdm import tqdm

import aerosandbox.numpy as np
from aerosandbox.tools.pretty_plots.utilities.natural_univariate_spline import NaturalUnivariateSpline as Spline
from scipy import signal
from aerosandbox.tools.code_benchmarking import Timer


def estimate_noise_standard_deviation(
        data: np.ndarray,
        estimator_order: int = None,
) -> float:
    """
    Estimates the standard deviation of the random noise in a time-series dataset.

    Relies on several assumptions:

    - The noise is normally-distributed and independent between samples (i.e. white noise).

    - The noise is stationary and homoscedastic (i.e., the noise standard deviation is constant).

    - The noise is uncorrelated with the signal.

    - The sample rate of the data is significantly higher than the highest-frequency component of the signal. (In
        practice, this ratio need not be more than ~5:1, if higher-order estimators are used. At a minimum, however,
        this ratio must be greater than 2:1, corresponding to the Nyquist frequency.)

    The algorithm used in this function is a highly-optimized version of the math described in this repository,
    part of an upcoming paper: https://github.com/peterdsharpe/aircraft-polar-reconstruction-from-flight-test

    Args:

        data: A 1D NumPy array of time-series data.

        estimator_order: The order of the estimator to use. Higher orders are generally more accurate, up to the
            point where sample error starts to dominate. If None, a reasonable estimator order will be chosen automatically.

    Returns: An estimate of the standard deviation of the data's noise component.

    """
    if len(data) < 2:
        raise ValueError("Data must have at least 2 points.")

    if estimator_order is None:
        estimator_order = int(np.clip(len(data) ** 0.5, 1, 1000))

    ##### Noise Variance Reconstruction #####
    from scipy.special import gammaln
    ln_factorial = lambda x: gammaln(x + 1)

    ### For speed, pre-compute the log-factorial of integers from 1 to estimator_order
    # ln_f = ln_factorial(np.arange(estimator_order + 1))
    ln_f = np.cumsum(
        np.log(
            np.concatenate([
                [1],
                np.arange(1, estimator_order + 1)
            ])
        )
    )

    ### Create a convolutional kernel to vectorize the summation
    log_coeffs = (
            2 * ln_f[estimator_order] - ln_f - ln_f[::-1] - 0.5 * ln_factorial(2 * estimator_order)
    )
    indices = np.nonzero(log_coeffs >= np.log(1e-20) + log_coeffs[estimator_order // 2])[0]
    coefficients = np.exp(log_coeffs[indices[0]:indices[-1] + 1])
    coefficients[::2] *= -1  # Flip the sign on every other coefficient
    coefficients -= np.mean(coefficients)  # Remove any bias introduced by floating-point error

    # sample_stdev = signal.convolve(data, coefficients[::-1], 'valid')
    sample_stdev = signal.oaconvolve(data, coefficients[::-1], 'valid')
    return np.mean(sample_stdev ** 2) ** 0.5


def bootstrap_fits(
        x: np.ndarray,
        y: np.ndarray,
        x_noise_stdev: Union[None, float] = 0.,
        y_noise_stdev: Union[None, float] = None,
        n_bootstraps: int = 2000,
        fit_points: Union[int, Iterable[float], None] = 300,
        spline_degree: int = 3,
        normalize: bool = None,
) -> Union[Tuple[np.ndarray, np.ndarray], List[Spline]]:
    """
    Bootstraps a time-series dataset and fits splines to each bootstrap resample.

    Args:

        x: The independent variable (e.g., time) of the dataset. A 1D NumPy array.

        y: The dependent variable (e.g., altitude) of the dataset. A 1D NumPy array.

        n_bootstraps: The number of bootstrap resamples to create.

        fit_points: An optional variable that determines what to do with the splines after they are fit:

            - If an integer, the splines will be evaluated at a linearly-spaced vector of points between the minimum
                and maximum x-values of the dataset, with the number of points equal to `fit_points`. This is the default.

            - If an iterable of floats (e.g. a 1D NumPy array), the splines will be evaluated at those points.

            - If None, the splines won't be evaluated, and instead the splines are returned directly.

        spline_degree: The degree of the splines to fit.

        normalize: Whether or not to normalize the data before fitting. If True, the data will be normalized to
            the range [0, 1] before fitting, and the splines will be un-normalized before being returned. If False,
            the data will not be normalized before fitting.

            - If None (the default), the data will be normalized if and only if `fit_points` is not None.


    Returns: One of the following, depending on the value of `fit_points`:

        - If `fit_points` is an integer or array, then this function returns a tuple of NumPy arrays:

            - `x_fit`: A 1D NumPy array with the x-values at which the splines were evaluated.

            - `y_bootstrap_fits`: A 2D NumPy array of shape (n_bootstraps, len(x_fit)) with the y-values of the
                splines evaluated at each bootstrap resample and at each x-value.

        - If `fit_points` is None, then this function returns a list of `n_bootstraps` splines, each of which is a
            `NaturalUnivariateSpline`, which is a subclass of `scipy.interpolate.UnivariateSpline` with more sensible
            extrapolation.

    """
    ### Set defaults
    if normalize is None:
        normalize = fit_points is not None

    ### Discard any NaN points
    isnan = np.logical_or(
        np.isnan(x),
        np.isnan(y),
    )
    x = x[~isnan]
    y = y[~isnan]

    ### Compute the standard deviation of the noise
    if x_noise_stdev is None:
        x_noise_stdev = estimate_noise_standard_deviation(x)
        print(f"Estimated x-component of noise standard deviation: {x_noise_stdev}")
    if y_noise_stdev is None:
        y_noise_stdev = estimate_noise_standard_deviation(y)
        print(f"Estimated y-component of noise standard deviation: {y_noise_stdev}")

    ### Sort the data by x-value
    sort_indices = np.argsort(x)
    x = x[sort_indices]
    y = y[sort_indices]

    ### Prepare for normalization
    x_min = np.min(x)
    x_max = np.max(x)
    x_rng = x_max - x_min

    y_min = np.min(y)
    y_max = np.max(y)
    y_rng = y_max - y_min

    if normalize:
        x_normalize = lambda x: (x - x_min) / x_rng
        y_normalize = lambda y: (y - y_min) / y_rng
        # x_unnormalize = lambda x_n: x_n * x_rng + x_min
        y_unnormalize = lambda y_n: y_n * y_rng + y_min

        x_stdev_normalized = x_noise_stdev / x_rng
        y_stdev_normalized = y_noise_stdev / y_rng

    else:
        x_normalize = lambda x: x
        y_normalize = lambda y: y
        # x_unnormalize = lambda x_n: x_n
        y_unnormalize = lambda y_n: y_n

        x_stdev_normalized = x_noise_stdev
        y_stdev_normalized = y_noise_stdev

    with tqdm(total=n_bootstraps, desc="Bootstrapping", unit=" samples") as progress_bar:
        splines = []
        n_valid_splines = 0
        n_attempted_splines = 0

        while n_valid_splines < n_bootstraps:

            n_attempted_splines += 1

            ### Obtain a bootstrap resample
            indices = np.random.choice(len(x), size=len(x), replace=True)

            x_sample = x[indices] + np.random.normal(scale=x_noise_stdev, size=len(x))
            y_sample = y[indices]

            order = np.argsort(x_sample)
            x_sample = x_sample[order]
            y_sample = y_sample[order]

            with warnings.catch_warnings():
                warnings.simplefilter("ignore", category=UserWarning)
                spline = Spline(
                    x=x_normalize(x_sample),
                    y=y_normalize(y_sample),
                    w=np.ones_like(x) / y_stdev_normalized,
                    s=len(x),
                    k=spline_degree,
                )

            if not np.isnan(spline(x_normalize((x_min + x_max) / 2))):
                n_valid_splines += 1
                progress_bar.update(1)
                splines.append(spline)
            else:
                continue

    if fit_points is None:
        return splines

    else:
        ### Determine which x-points to resample at
        if fit_points is None:
            x_fit = None
            if normalize:
                raise ValueError("If `fit_points` is None, `normalize` must be False.")
        elif isinstance(fit_points, int):
            x_fit = np.linspace(
                np.min(x),
                np.max(x),
                fit_points
            )
        else:
            x_fit = np.array(fit_points)

        ### Evaluate the splines at the x-points
        y_bootstrap_fits = np.array([
            y_unnormalize(spline(x_normalize(x_fit)))
            for spline in splines
        ])

        ### Throw an error if all of the splines are NaN
        if np.all(np.isnan(y_bootstrap_fits)):
            raise ValueError("All of the splines are NaN. This is likely due to a poor choice of `spline_degree`.")

        return x_fit, y_bootstrap_fits


if __name__ == '__main__':
    np.random.seed(1)
    N = 1000
    f_sample_over_f_signal = 1000

    t = np.arange(N)
    y = np.sin(2 * np.pi / f_sample_over_f_signal * t) + 0.1 * np.random.randn(len(t))

    print(estimate_noise_standard_deviation(y, 1))

    # d = dict(np.load("raw_data.npz"))
    #
    # x = d["airspeed"]
    # y = d["voltage"] * d["current"]
    #
    # # estimate_noise_standard_deviation(x)
    # #
    # # x_fit, y_bootstrap_fits = bootstrap_fits(
    # #     x, y,
    # #     x_stdev=None,
    # #     y_stdev=None,
    # #     n_bootstraps=20,
    # #     spline_degree=5,
    # # )
    # import matplotlib.pyplot as plt
    # import aerosandbox.tools.pretty_plots as p
    #
    # fig, ax = plt.subplots(figsize=(7, 4))
    #
    # p.plot_with_bootstrapped_uncertainty(
    #     x, y,
    #     x_stdev=None,
    #     y_stdev=estimate_noise_standard_deviation(y[np.argsort(x)]),
    #     ci=[0.75, 0.95],
    #     color="coral",
    #     n_bootstraps=100,
    #     n_fit_points=200,
    #     # ci_to_alpha_mapping=lambda ci: 0.4,
    #     normalize=False,
    #     spline_degree=3,
    # )
    # plt.xlim(x.min(), x.max())
    # plt.ylim(-10, 800)
    # p.set_ticks(1, 0.25, 100, 25)
    # plt.legend(
    #     loc="lower right"
    # )
    # p.show_plot(
    #     xlabel="Cruise Airspeed [m/s]",
    #     ylabel="Electrical Power Required [W]",
    #     title="Raw Data",
    #     legend=False,
    #     dpi=300
    # )



import aerosandbox.numpy as np

np.random.seed(0)  # Fix a seed for reproducibility

### Create some data (some fictional system where temperature is a function of time, and we're measuring it)
n = 20

time = 100 * np.random.rand(n)

actual_temperature = 2 * time + 20  # True physics of the system

noise = 10 * np.random.randn(n)
measured_temperature = actual_temperature + noise  # Measured temperature of the system

### Add in a dropout measurement (say, the sensor wire randomly came loose and gave us a 0 reading)
time = np.hstack((time, 90))
measured_temperature = np.hstack((measured_temperature, 0))

if __name__ == '__main__':
    from aerosandbox.tools.pretty_plots import plt, sns, mpl, show_plot

    fig, ax = plt.subplots()
    plt.plot(time, measured_temperature, ".")
    show_plot(
        xlabel="Time",
        ylabel="Measured Temperature"
    )



import aerosandbox as asb
import aerosandbox.numpy as np
import sympy as s
from aerosandbox.numpy.integrate_discrete import integrate_discrete_squared_curvature

n_samples = 10000

x = s.symbols("x", real=True)
k = s.symbols("k", positive=True, real=True)
f = s.cos(k * x * 2 * s.pi) / k ** 2
d2fdx2 = f.diff(x, 2)
exact = s.simplify(s.integrate(d2fdx2 ** 2, (x, 0, 1)))


@np.vectorize
def get_approx(period=10, method="cubic"):
    x_vals = np.linspace(0, 1, n_samples).astype(float)
    f_vals = s.lambdify(
        x,
        f.subs(k, (n_samples - 1) / period),
    )(x_vals)

    approx = np.sum(integrate_discrete_squared_curvature(
        f=f_vals,
        x=x_vals,
        method=method,
    ))

    return approx


periods = np.geomspace(2, n_samples, 1001)
exacts = s.lambdify(k, exact)((n_samples - 1) / periods)

import matplotlib.pyplot as plt
import aerosandbox.tools.pretty_plots as p

fig, ax = plt.subplots(2, 1, figsize=(6, 8))
for method in ["cubic", "simpson", "hybrid_simpson_cubic"]:
    approxes = get_approx(periods, method)
    ratio = approxes / exacts
    rel_errors = np.abs(approxes - exacts) / exacts
    ax[0].loglog(periods, rel_errors, ".-", label=method, alpha=0.8, markersize=3)
    ax[1].semilogx(periods, ratio, ".-", label=method, alpha=0.8, markersize=3)
plt.xlabel("Period [samples]")
ax[0].set_ylabel("Relative Error")
ax[1].set_ylabel("Approximation / Exact")
plt.sca(ax[1])
ax[1].set_ylim(bottom=0)
ax[1].plot(
    periods,
    np.ones_like(periods),
    label="Exact",
    color="k",
    linestyle="--",
    alpha=0.5
)
p.show_plot()



import matplotlib.pyplot as plt
import matplotlib.transforms as transforms

def hline(
        y,
        linestyle="--",
        color="k",
        text: str = None,
        text_xloc=0.5,
        text_ha="center",
        text_va="bottom",
        text_kwargs=None,
        **kwargs
):  # TODO docs
    if text_kwargs is None:
        text_kwargs = {}
    ax = plt.gca()
    plt.axhline(y=y, ls=linestyle, color=color, **kwargs)
    if text is not None:
        trans = transforms.blended_transform_factory(
            ax.transAxes, ax.transData
        )
        plt.annotate(
            text=text,
            xy=(text_xloc, y),
            xytext=(0, 0),
            xycoords=trans,
            textcoords="offset points",
            ha=text_ha,
            va=text_va,
            color=color,
            **text_kwargs
        )


def vline(
        x,
        linestyle="--",
        color="k",
        text: str = None,
        text_yloc=0.5,
        text_ha="right",
        text_va="center",
        text_kwargs=None,
        **kwargs
):  # TODO docs
    if text_kwargs is None:
        text_kwargs = {}
    ax = plt.gca()
    plt.axvline(x=x, ls=linestyle, color=color, **kwargs)
    if text is not None:
        trans = transforms.blended_transform_factory(
            ax.transData, ax.transAxes
        )
        plt.annotate(
            text=text,
            xy=(x, text_yloc),
            xytext=(0, 0),
            xycoords=trans,
            textcoords="offset points",
            ha=text_ha,
            va=text_va,
            color=color,
            rotation=90,
            **text_kwargs
        )


import hashlib
import aerosandbox.numpy as np


def eng_string(
        x: float,
        unit: str = "",
        format='%.3g',
        si=True,
        add_space_after_number: bool = None,
) -> str:
    '''
    Taken from: https://stackoverflow.com/questions/17973278/python-decimal-engineering-notation-for-mili-10e-3-and-micro-10e-6/40691220

    Returns float/int value <x> formatted in a simplified engineering format -
    using an exponent that is a multiple of 3.

    Args:

        x: The value to be formatted. Float or int.

        unit: A unit of the quantity to be expressed, given as a string. Example: Newtons -> "N"

        format: A printf-style string used to format the value before the exponent.

        si: if true, use SI suffix for exponent. (k instead of e3, n instead of
            e-9, etc.)

    Examples:

    With format='%.2f':
        1.23e-08 -> 12.30e-9
             123 -> 123.00
          1230.0 -> 1.23e3
      -1230000.0 -> -1.23e6

    With si=True:
          1230.0 -> "1.23k"
      -1230000.0 -> "-1.23M"

    With unit="N" and si=True:
          1230.0 -> "1.23 kN"
      -1230000.0 -> "-1.23 MN"
    '''

    sign = ''
    if x < 0:
        x = -x
        sign = '-'
    elif x == 0:
        return format % 0
    elif np.isnan(x):
        return "NaN"

    exp = int(np.floor(np.log10(x)))
    exp3 = exp - (exp % 3)
    x3 = x / (10 ** exp3)

    if si and exp3 >= -24 and exp3 <= 24:
        if exp3 == 0:
            suffix = ""
        else:
            suffix = 'yzafpnμm kMGTPEZY'[(exp3 + 24) // 3]

        if add_space_after_number is None:
            add_space_after_number = (unit != "")

        if add_space_after_number:
            suffix = " " + suffix + unit
        else:
            suffix = suffix + unit

    else:
        suffix = f'e{exp3}'

        if add_space_after_number:
            add_space_after_number = (unit != "")

        if add_space_after_number:
            suffix = suffix + " " + unit
        else:
            suffix = suffix + unit

    return f"{sign}{format % x3}{suffix}"


def latex_sci_notation_string(
        x: float,
        format='%.2e',
) -> str:
    """
    Converts a floating-point number to a LaTeX-style formatted string. Does not include the `$$` wrapping to put you in math mode.

    Does not use scientific notation if the base would be zero.

    Examples:

        latex_sci_notation_string(3000) -> '3 \\times 10^{3}'
    """
    float_str = format % x
    base, exponent = float_str.split("e")
    exponent = int(exponent)
    if exponent == 0:
        return base
    else:
        return r"{0} \times 10^{{{1}}}".format(base, exponent)


def hash_string(string: str) -> int:
    """
    Hashes a string into a quasi-random 32-bit integer! (Based on an MD5 checksum algorithm.)

    Usual warnings apply: it's MD5, don't use this for anything intended to be cryptographically secure.
    """
    md5 = hashlib.md5(string.encode('utf-8'))
    hash_hex = md5.hexdigest()
    hash_int = int(hash_hex, 16)
    hash_int64 = hash_int % (2 ** 32)
    return hash_int64


def trim_string(string: str, length: int = 80) -> str:
    """
    Trims a string to be less than a given length. If the string would exceed the length, makes it end in ellipses ("…").

    Args:

        string: The string to be trimmed.

        length: The length to trim the string to, including any ellipses that may be added.

    Returns: The trimmed string, including ellipses if needed.

    """
    if len(string) > length:
        return string[:length - 1] + "…"
    else:
        return string


def has_balanced_parentheses(string: str, left="(", right=")") -> bool:
    """
    Determines whether a string has matching parentheses or not.

    Examples:

        >>> has_balanced_parentheses("3 * (x + (2 ** 5))") -> True

        >>> has_balanced_parentheses("3 * (x + (2 ** 5)") -> False

    Args:

        string: The string to be evaluated.

        left: The left parentheses. Can be modified if, for example, you need to check square brackets.

        right: The right parentheses. Can be modified if, for example, you need to check square brackets.

    Returns: A boolean of whether or not the string has balanced parentheses.

    """
    parenthesis_level = 0

    for char in string:
        if char == left:
            parenthesis_level += 1
        elif char == right:
            parenthesis_level -= 1

    return parenthesis_level == 0


def wrap_text_ignoring_mathtext(
        text: str,
        width: int = 70,
) -> str:
    r"""
    Reformat the single paragraph in 'text' to fit in lines of no more
    than 'width' columns, and return a new string containing the entire
    wrapped paragraph.  Tabs are expanded and other
    whitespace characters converted to space.

    Similar to `textwrap.fill`, but keeps any mathtext blocks contiguous and unaltered. Mathtext blocks are segments of `text` that are between $ markers, to indicate LaTeX-like formatting. Dollar-sign literals (\$) do not trigger Mathtext, and that is respected here as well.

    For example:
        >>> wrap_text_ignoring_mathtext()

    Args:

        text: The text to be wrapped.

        width: The maximum width of wrapped lines (unless break_long_words is false)

    Returns:

        A string containing the entire paragraph with line breaks as newline ("\n") characters.

    """
    import textwrap, re

    # Pattern to match mathtext blocks
    mathtext_trigger = r"(?<!\\)(?:\\\\)*\$"

    # Split the text into non-mathtext parts and mathtext parts
    parts = re.split(mathtext_trigger, text)
    text_parts = [part for i, part in enumerate(parts) if i % 2 == 0]
    math_parts = [part for i, part in enumerate(parts) if i % 2 == 1]

    # Reassemble the result
    output = ""
    cursor_position = 0

    while len(text_parts) + len(math_parts) > 0:
        try:
            text_part = text_parts.pop(0)

            contribution = textwrap.fill(
                text_part,
                width=width,
                initial_indent=" " * cursor_position,
                drop_whitespace=False,
                break_on_hyphens=True,
                break_long_words=False,
            )[cursor_position:]

            output += contribution

            if "\n" in contribution:
                cursor_position = len(contribution.split("\n")[-1])
            else:
                cursor_position += len(contribution)

        except IndexError:
            pass

        try:
            math_part = math_parts.pop(0)

            estimated_space: int = int(np.round(len(math_part) * 0.5))

            if cursor_position + estimated_space < width:
                output += f"${math_part}$"
                cursor_position += estimated_space
            else:
                output += f"\n${math_part}$"
                cursor_position = estimated_space

        except IndexError:
            pass

    output = "\n".join([line.strip() for line in output.split("\n")])

    return output


if __name__ == '__main__':
    for input in [
        r"$ax^2+bx+c$",
        r"Photon flux $\phi$",
        r"Photon flux $\phi$ is given by $\phi = \frac{c}{\lambda}$",
        r"Earnings for 2022 $M\$/year$",
        r"$ax^2+bx+c$ and also $3x$"
    ]:
        print(wrap_text_ignoring_mathtext(input, width=10))

    print(
        wrap_text_ignoring_mathtext(
            r"Angle of Attack $\alpha$ [deg]",
            14
        )
    )


import aerosandbox as asb
import aerosandbox.numpy as np
from aerosandbox.numpy.integrate import solve_ivp
import pytest


def test_racecar_adaptive_ode(plot=False):

    N = 100  # number of control intervals

    opti = asb.Opti()  # Optimization problem

    t_final = opti.variable(init_guess=1, lower_bound=1e-2)
    u_amp = opti.variable(init_guess=np.ones(10), lower_bound=0, upper_bound=1)

    # u_amp = opti.variable(init_guess=np.ones(20) * 0.5, lower_bound=-0.1, upper_bound=0.1)

    # u_amp = opti.variable(init_guess=np.eye(10)[0],)

    def u(t):
        u = 0
        # # Fourier
        # for i in range(np.length(u_amp)):
        #     u += u_amp[i] * np.cos(np.pi * (i) * t / t_final)

        # Bernstein
        from scipy.special import comb
        tn = t / t_final
        for i in range(np.length(u_amp)):
            u += (
                    u_amp[i]
                    * comb(np.length(u_amp) - 1, i)
                    * tn ** i
                    * (1 - tn) ** (np.length(u_amp) - 1 - i)
            )

        # # # Step
        # index = np.floor(t / t_final * np.length(u_amp))
        # for i in range(np.length(u_amp)):
        #     if i == 0:
        #         weight = np.where(index <= 0, 1, 0)
        #     elif i == np.length(u_amp) - 1:
        #         weight = np.where(index >= i, 1, 0)
        #     else:
        #         weight = np.where(index == i, 1, 0)
        #
        #     u += u_amp[i] * weight

        return u

    def func(t, y):
        return np.array([
            y[1],
            u(t) - y[1],
        ])

    res = solve_ivp(
        fun=func,
        t_span=(0, t_final),
        y0=[0, 0],
    )
    t = res.t
    x = res.y[0, :]
    v = res.y[1, :]

    opti.minimize(t_final)

    opti.subject_to([
        v <= 1 - np.sin(2 * np.pi * x) / 2,
        x[-1] >= 1,
    ])

    def callback(i):
        soli = asb.OptiSol(opti, opti.debug)

        print(f"t_final: {soli(t_final)}")

        if i % 5 == 0:
            import matplotlib.pyplot as plt
            import aerosandbox.tools.pretty_plots as p
            fig, ax = plt.subplots()
            ax.plot(soli(t), soli(v), label="speed")
            ax.plot(soli(t), soli(x), label="position")
            ax.plot(soli(t), soli(1 - np.sin(2 * np.pi * x) / 2), "r--", label="speed limit")
            t_plot = np.linspace(0, soli(t_final), 1000)
            ax.plot(t_plot, soli(u(t_plot)), "k", label="throttle")
            plt.ylim(-0.1, 1.6)
            p.show_plot(f"i = {i}")

    sol = opti.solve(
        # callback=callback,
        # verbose=False,
        # behavior_on_failure="return_last"
    )

    assert sol(t_final) == pytest.approx(1.9683594, abs=1e-3)

    if plot:
        import matplotlib.pyplot as plt
        import aerosandbox.tools.pretty_plots as p

        fig, ax = plt.subplots()
        ax.plot(sol(t), sol(v), label="speed")
        ax.plot(sol(t), sol(x), label="position")
        ax.plot(sol(t), sol(1 - np.sin(2 * np.pi * x) / 2), "r--", label="speed limit")
        t_plot = np.linspace(0, sol(t_final), 1000)
        ax.plot(t_plot, sol(u(t_plot)), "k", label="throttle")
        plt.xlabel("Time [s]")
        plt.ylabel("Position [m] / Speed [m/s] / Throttle [-]")
        p.show_plot(
            rotate_axis_labels=False
        )



import aerosandbox.numpy as np


def haack_series(
        x_over_L: np.ndarray,
        C=1 / 3
):
    theta = np.arccos(1 - 2 * x_over_L)
    radius = ((
                      theta - np.sin(2 * theta) / 2 + C * np.sin(theta) ** 3
              ) / np.pi) ** 0.5
    return radius


def karman(
        x_over_L: np.ndarray,
):
    return haack_series(
        x_over_L=x_over_L,
        C=0
    )


def LV_haack(
        x_over_L: np.ndarray,
):
    return haack_series(
        x_over_L=x_over_L,
        C=1 / 3
    )


def tangent(
        x_over_L: np.ndarray,
):
    return haack_series(
        x_over_L=x_over_L,
        C=2 / 3
    )


if __name__ == '__main__':
    import matplotlib.pyplot as plt
    import aerosandbox.tools.pretty_plots as p

    fig, ax = plt.subplots()

    x_over_L = np.cosspace(0, 1, 2000)
    FR = 3

    plt.plot(x_over_L * FR, karman(x_over_L), label="$C=0$ (LD-Haack, Karman)")
    plt.plot(x_over_L * FR, LV_haack(x_over_L), label="$C=1/3$ (LV-Haack)")
    plt.plot(x_over_L * FR, tangent(x_over_L), label="$C=2/3$ (Tangent)")

    p.equal()
    p.show_plot(
        f"Nosecone Haack Series\nFineness Ratio $FR = {FR}$",
        "Length $x$",
        "Radius $r$"
    )



import aerosandbox.numpy as np
from typing import Union


def calculate_induced_velocity_horseshoe(
        x_field: Union[float, np.ndarray],
        y_field: Union[float, np.ndarray],
        z_field: Union[float, np.ndarray],
        x_left: Union[float, np.ndarray],
        y_left: Union[float, np.ndarray],
        z_left: Union[float, np.ndarray],
        x_right: Union[float, np.ndarray],
        y_right: Union[float, np.ndarray],
        z_right: Union[float, np.ndarray],
        gamma: Union[float, np.ndarray] = 1,
        trailing_vortex_direction: np.ndarray = None,
        vortex_core_radius: float = 0,
) -> [Union[float, np.ndarray], Union[float, np.ndarray], Union[float, np.ndarray]]:
    """
    Calculates the induced velocity at a point:
        [x_field, y_field, z_field]
    in a 3D potential-flow flowfield.

    In this flowfield, the following singularity elements are assumed:
        * A single horseshoe vortex consisting of a bound leg and two trailing legs

    This function consists entirely of scalar, elementwise NumPy ufunc operations - so it can be vectorized as
    desired assuming input dimensions/broadcasting are compatible.

    Args:
        x_field: x-coordinate of the field point

        y_field: y-coordinate of the field point

        z_field: z-coordinate of the field point

        x_left: x-coordinate of the left vertex of the bound vortex

        y_left: y-coordinate of the left vertex of the bound vortex

        z_left: z-coordinate of the left vertex of the bound vortex

        x_right: x-coordinate of the right vertex of the bound vortex

        y_right: y-coordinate of the right vertex of the bound vortex

        z_right: z-coordinate of the right vertex of the bound vortex

        gamma: The strength of the horseshoe vortex filament.

        trailing_vortex_direction: The direction that the trailing legs of the horseshoe vortex extend. Usually,
        this is modeled as the direction of the freestream.

        vortex_core_radius: To prevent a vortex singularity, here we use a Kaufmann vortex model. This parameter
        governs the radius of this vortex model. It should be significantly smaller (e.g., at least an order of
        magnitude smaller) than the smallest bound leg in the analysis in question.

    Returns: u, v, and w:
        The x-, y-, and z-direction induced velocities.
    """
    if trailing_vortex_direction is None:
        trailing_vortex_direction = np.array([1, 0, 0])

    np.assert_equal_shape({
        "x_field": x_field,
        "y_field": y_field,
        "z_field": z_field,
    })
    np.assert_equal_shape({
        "x_left" : x_left,
        "y_left" : y_left,
        "z_left" : z_left,
        "x_right": x_right,
        "y_right": y_right,
        "z_right": z_right,
    })

    a_x = np.add(x_field, -x_left)
    a_y = np.add(y_field, -y_left)
    a_z = np.add(z_field, -z_left)

    b_x = np.add(x_field, -x_right)
    b_y = np.add(y_field, -y_right)
    b_z = np.add(z_field, -z_right)

    u_x = trailing_vortex_direction[0]
    u_y = trailing_vortex_direction[1]
    u_z = trailing_vortex_direction[2]

    # Handle the special case where the field point is on one of the legs (either bound or trailing)
    def smoothed_inv(x):
        "Approximates 1/x with a function that sharply goes to 0 in the x -> 0 limit."
        if not np.all(vortex_core_radius == 0):
            return x / (x ** 2 + vortex_core_radius ** 2)
        else:
            return 1 / x

    ### Do some useful arithmetic

    a_cross_b_x = a_y * b_z - a_z * b_y
    a_cross_b_y = a_z * b_x - a_x * b_z
    a_cross_b_z = a_x * b_y - a_y * b_x
    a_dot_b = a_x * b_x + a_y * b_y + a_z * b_z

    a_cross_u_x = a_y * u_z - a_z * u_y
    a_cross_u_y = a_z * u_x - a_x * u_z
    a_cross_u_z = a_x * u_y - a_y * u_x
    a_dot_u = a_x * u_x + a_y * u_y + a_z * u_z

    b_cross_u_x = b_y * u_z - b_z * u_y
    b_cross_u_y = b_z * u_x - b_x * u_z
    b_cross_u_z = b_x * u_y - b_y * u_x
    b_dot_u = b_x * u_x + b_y * u_y + b_z * u_z

    norm_a = (a_x ** 2 + a_y ** 2 + a_z ** 2) ** 0.5
    norm_b = (b_x ** 2 + b_y ** 2 + b_z ** 2) ** 0.5
    norm_a_inv = smoothed_inv(norm_a)
    norm_b_inv = smoothed_inv(norm_b)

    ### Calculate Vij

    term1 = (norm_a_inv + norm_b_inv) * smoothed_inv(norm_a * norm_b + a_dot_b)
    term2 = norm_a_inv * smoothed_inv(norm_a - a_dot_u)
    term3 = norm_b_inv * smoothed_inv(norm_b - b_dot_u)

    constant = gamma / (4 * np.pi)

    u = np.multiply(
        constant,
        (
                a_cross_b_x * term1 +
                a_cross_u_x * term2 -
                b_cross_u_x * term3
        )
    )
    v = np.multiply(
        constant,
        (
                a_cross_b_y * term1 +
                a_cross_u_y * term2 -
                b_cross_u_y * term3
        )
    )
    w = np.multiply(
        constant,
        (
                a_cross_b_z * term1 +
                a_cross_u_z * term2 -
                b_cross_u_z * term3
        )
    )

    return u, v, w


if __name__ == '__main__':
    ##### Check single vortex
    u, v, w = calculate_induced_velocity_horseshoe(
        x_field=0,
        y_field=0,
        z_field=0,
        x_left=-1,
        y_left=-1,
        z_left=0,
        x_right=-1,
        y_right=1,
        z_right=0,
        gamma=1,
    )
    print(u, v, w)

    ##### Plot grid of single vortex
    args = (-2, 2, 30)
    x = np.linspace(*args)
    y = np.linspace(*args)
    z = np.linspace(*args)
    X, Y, Z = np.meshgrid(x, y, z)

    Xf = X.flatten()
    Yf = Y.flatten()
    Zf = Z.flatten()

    left = [0, -1, 0]
    right = [0, 1, 0]

    Uf, Vf, Wf = calculate_induced_velocity_horseshoe(
        x_field=Xf,
        y_field=Yf,
        z_field=Zf,
        x_left=left[0],
        y_left=left[1],
        z_left=left[2],
        x_right=right[0],
        y_right=right[1],
        z_right=right[2],
        gamma=1,
    )

    pos = np.stack((Xf, Yf, Zf)).T
    dir = np.stack((Uf, Vf, Wf)).T

    dir_norm = np.reshape(np.linalg.norm(dir, axis=1), (-1, 1))

    dir = dir / dir_norm * dir_norm ** 0.2

    import pyvista as pv

    pv.set_plot_theme('dark')
    plotter = pv.Plotter()
    plotter.add_arrows(
        cent=pos,
        direction=dir,
        mag=0.15
    )
    plotter.add_lines(
        lines=np.array([
            [Xf.max(), left[1], left[2]],
            left,
            right,
            [Xf.max(), right[1], right[2]]
        ])
    )
    plotter.show_grid()
    plotter.show()

    ##### Check multiple vortices
    args = (-2, 2, 30)
    x = np.linspace(*args)
    y = np.linspace(*args)
    z = np.linspace(*args)
    X, Y, Z = np.meshgrid(x, y, z)

    Xf = X.flatten()
    Yf = Y.flatten()
    Zf = Z.flatten()

    left = [0, -1, 0]
    center = [0, 0, 0]
    right = [0, 1, 0]

    lefts = np.array([left, center])
    rights = np.array([center, right])
    strengths = np.array([2, 1])


    def wide(array):
        return np.reshape(array, (1, -1))


    def tall(array):
        return np.reshape(array, (-1, 1))


    Uf_each, Vf_each, Wf_each = calculate_induced_velocity_horseshoe(
        x_field=wide(Xf),
        y_field=wide(Yf),
        z_field=wide(Zf),
        x_left=tall(lefts[:, 0]),
        y_left=tall(lefts[:, 1]),
        z_left=tall(lefts[:, 2]),
        x_right=tall(rights[:, 0]),
        y_right=tall(rights[:, 1]),
        z_right=tall(rights[:, 2]),
        gamma=tall(strengths),
    )

    Uf = np.sum(Uf_each, axis=0)
    Vf = np.sum(Vf_each, axis=0)
    Wf = np.sum(Wf_each, axis=0)

    pos = np.stack((Xf, Yf, Zf)).T
    dir = np.stack((Uf, Vf, Wf)).T

    dir_norm = np.reshape(np.linalg.norm(dir, axis=1), (-1, 1))

    dir = dir / dir_norm * dir_norm ** 0.2

    import pyvista as pv

    pv.set_plot_theme('dark')
    plotter = pv.Plotter()
    plotter.add_arrows(
        cent=pos,
        direction=dir,
        mag=0.15
    )
    plotter.add_lines(
        lines=np.array([
            [Xf.max(), left[1], left[2]],
            left,
            center,
            [Xf.max(), center[1], center[2]],
            center,
            right,
            [Xf.max(), right[1], right[2]]
        ])
    )
    plotter.show_grid()
    plotter.show()
